# This file is part of Xpra.
# Copyright (C) 2012 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import re
from time import monotonic
from typing import Any
from collections.abc import Sequence, Callable
from ctypes import create_unicode_buffer, sizeof, byref, c_ulong
from ctypes.wintypes import RECT, POINT, BYTE

from xpra.util.screen import prettify_plug_name
from xpra.util.str_fn import csv
from xpra.util.env import envbool
from xpra.util.system import is_VirtualBox
from xpra.common import XPRA_APP_ID, NotificationID
from xpra.scripts.config import InitException
from xpra.server.gtk_server import GTKServerBase
from xpra.server.shadow.shadow_server_base import try_setup_capture
from xpra.server.shadow.gtk_root_window_model import GTKImageCapture
from xpra.server.shadow.gtk_shadow_server_base import GTKShadowServerBase
from xpra.server.shadow.root_window_model import RootWindowModel
from xpra.platform.win32 import constants as win32con
from xpra.platform.win32.gui import get_desktop_name, get_fixed_cursor_size
from xpra.platform.win32.keyboard_config import KeyboardConfig, fake_key
from xpra.platform.win32.events import get_win32_event_listener, POWER_EVENTS
from xpra.platform.win32.shadow_cursor import get_cursor_data
from xpra.platform.win32.gdi_screen_capture import GDICapture
from xpra.log import Logger

# user32:
from xpra.platform.win32.common import (
    EnumWindows, EnumWindowsProc, FindWindowA, IsWindowVisible,
    GetWindowTextLengthW, GetWindowTextW,
    GetWindowRect,
    GetWindowThreadProcessId,
    GetSystemMetrics,
    SetPhysicalCursorPos, GetPhysicalCursorPos, GetCursorInfo, CURSORINFO,
    GetKeyboardState, SetKeyboardState,
    EnumDisplayMonitors, GetMonitorInfo,
    mouse_event,
)

log = Logger("shadow", "win32")
shapelog = Logger("shape")
cursorlog = Logger("cursor")
keylog = Logger("keyboard")
screenlog = Logger("screen")


SEAMLESS = envbool("XPRA_WIN32_SEAMLESS", False)
NVFBC = envbool("XPRA_SHADOW_NVFBC", True)
GDI = envbool("XPRA_SHADOW_GDI", True)
GSTREAMER = envbool("XPRA_SHADOW_GSTREAMER", True)


def check_gstreamer_d3d11() -> bool:
    return not is_VirtualBox()


def check_gstreamer_dx9() -> bool:
    return True


def check_gstreamer_gdi() -> bool:
    # we don't want to be using GStreamer for GDI capture
    return False


def get_gstreamer_capture_elements() -> Sequence[str]:
    if "XPRA_GSTREAMER_CAPTURE_ELEMENTS" in os.environ:
        elements = os.environ.get("XPRA_GSTREAMER_CAPTURE_ELEMENTS", "").split(",")
    else:
        elements = []
        for element, check in {
            "d3d11screencapturesrc" : check_gstreamer_d3d11,
            "dx9screencapsrc": check_gstreamer_dx9,
            "gdiscreencapsrc": check_gstreamer_gdi,
        }.items():
            if envbool("XPRA_GSTREAMER_CAPTURE_"+element.upper(), False) or check():
                elements.append(element)
    log(f"get_gstreamer_capture_elements()={elements}")
    return elements


GSTREAMER_CAPTURE_ELEMENTS: Sequence[str] = get_gstreamer_capture_elements()


def check_gstreamer() -> bool:
    if not GSTREAMER:
        return False
    from xpra.gstreamer.common import has_plugins, import_gst
    import_gst()
    return any(has_plugins(element) for element in GSTREAMER_CAPTURE_ELEMENTS)


def check_nvfbc() -> bool:
    if not NVFBC:
        return False
    from xpra.codecs.nvidia.nvfbc.capture import get_capture_instance
    assert get_capture_instance
    return NVFBC


def check_gdi() -> bool:
    return GDI


def check_gtk() -> bool:
    from xpra.gtk import signals
    assert signals
    return True


SHADOW_OPTIONS = {
    "auto": lambda: True,
    "nvfbc": check_nvfbc,
    "gstreamer": check_gstreamer,
    "gdi": check_gdi,
    "gtk": check_gtk,
}


NOEVENT = (0, 0)
BUTTON_EVENTS: dict[tuple[int, bool], tuple[int, int]] = {
    # (button,up-or-down)  : win-event-name
    (1, True): (win32con.MOUSEEVENTF_LEFTDOWN, 0),
    (1, False): (win32con.MOUSEEVENTF_LEFTUP, 0),
    (2, True): (win32con.MOUSEEVENTF_MIDDLEDOWN, 0),
    (2, False): (win32con.MOUSEEVENTF_MIDDLEUP, 0),
    (3, True): (win32con.MOUSEEVENTF_RIGHTDOWN, 0),
    (3, False): (win32con.MOUSEEVENTF_RIGHTUP, 0),
    (4, True): (win32con.MOUSEEVENTF_WHEEL, win32con.WHEEL_DELTA),
    (4, False): NOEVENT,
    (5, True): (win32con.MOUSEEVENTF_WHEEL, -win32con.WHEEL_DELTA),
    (5, False): NOEVENT,
    (6, True): (win32con.MOUSEEVENTF_HWHEEL, win32con.WHEEL_DELTA),
    (6, False): NOEVENT,
    (7, True): (win32con.MOUSEEVENTF_HWHEEL, -win32con.WHEEL_DELTA),
    (7, False): NOEVENT,
    (8, True): (win32con.MOUSEEVENTF_XDOWN, win32con.XBUTTON1),
    (8, False): (win32con.MOUSEEVENTF_XUP, win32con.XBUTTON1),
    (9, True): (win32con.MOUSEEVENTF_XDOWN, win32con.XBUTTON2),
    (9, False): (win32con.MOUSEEVENTF_XUP, win32con.XBUTTON2),
}


def get_root_window_size() -> tuple[int, int]:
    w = GetSystemMetrics(win32con.SM_CXVIRTUALSCREEN)
    h = GetSystemMetrics(win32con.SM_CYVIRTUALSCREEN)
    return w, h


def get_monitors() -> list[dict[str, Any]]:
    monitors = []
    for m in EnumDisplayMonitors():
        mi = GetMonitorInfo(m)
        monitors.append(mi)
    return monitors


def setup_nvfbc_capture(w: int, h: int, pixel_depth=32):
    if not NVFBC:
        return None
    from xpra.codecs.nvidia.nvfbc.capture import get_capture_instance
    capture = get_capture_instance()
    try:
        pixel_format = {
            24: "RGB",
            32: "BGRX",
            30: "r210",
        }[pixel_depth]
        capture.init_context(w, h, pixel_format)
        # this will test the capture and ensure we can call get_image()
        capture.refresh()
        return capture
    except Exception as e:
        log("NvFBC_Capture", exc_info=True)
        log.warn("Warning: NvFBC screen capture initialization failed:")
        for x in str(e).replace(". ", ":").split(":"):
            if x.strip() and x != "nvfbc":
                log.warn(" %s", x.strip())
        return None


def setup_gstreamer_capture(w: int, h: int, pixel_depth=32):
    from xpra.codecs.gstreamer.capture import Capture
    for el in GSTREAMER_CAPTURE_ELEMENTS:
        log(f"testing gstreamer capture using {el}")
        try:
            capture = Capture(el, pixel_format="BGRX", width=w, height=h)
            capture.start()
            image = capture.get_image(0, 0, w, h)
            if image:
                log(f"using gstreamer element {el}")
                return capture
        except Exception:
            log(f"gstreamer failed to capture the screen using {el}", exc_info=True)
    return None


def setup_gdi_capture(w: int, h: int, pixel_depth=32):
    return GDICapture()


def setup_gtk_capture(w: int, h: int, pixel_depth=32):
    return GTKImageCapture(None)


CAPTURE_BACKENDS: dict[str, Callable] = {
    "nvfbc": setup_nvfbc_capture,
    "gstreamer": setup_gstreamer_capture,
    "gdi": setup_gdi_capture,
    "gtk": setup_gtk_capture,
}


class SeamlessRootWindowModel(RootWindowModel):

    def __init__(self, root, capture, title, geometry):
        super().__init__(root, capture, title, geometry)
        log("SeamlessRootWindowModel(%s, %s) SEAMLESS=%s", root, capture, SEAMLESS)
        self.property_names.append("shape")
        self.dynamic_property_names.append("shape")
        self.rectangles = self.get_shape_rectangles(logit=True)

    def refresh_shape(self) -> None:
        rectangles = self.get_shape_rectangles()
        if rectangles == self.rectangles:
            return  # unchanged
        self.rectangles = rectangles
        shapelog("refresh_shape() sending notify for updated rectangles: %s", rectangles)
        self.notify("shape")

    def get_shape_rectangles(self, logit=False) -> list:
        # get the list of windows
        log_fn = log.debug
        if logit or envbool("XPRA_SHAPE_DEBUG", False):
            log_fn = shapelog.debug
        taskbar = FindWindowA("Shell_TrayWnd", None)
        log_fn("taskbar window=%#x", taskbar)
        ourpid = os.getpid()
        log_fn("our pid=%i", ourpid)
        rectangles = []

        def enum_windows_cb(hwnd: int, lparam: int) -> bool:
            if not IsWindowVisible(hwnd):
                log_fn("skipped invisible window %#x", hwnd)
                return True
            pid = c_ulong()
            thread_id = GetWindowThreadProcessId(hwnd, byref(pid))
            if pid == ourpid:
                log_fn("skipped our own window %#x", hwnd)
                return True
            # skipping IsWindowEnabled check
            length = GetWindowTextLengthW(hwnd)
            buf = create_unicode_buffer(length + 1)
            if GetWindowTextW(hwnd, buf, length + 1) > 0:
                window_title = buf.value
            else:
                window_title = ''
            log_fn("get_shape_rectangles() found window '%s' with pid=%i and thread id=%i", window_title, pid, thread_id)
            rect = RECT()
            if GetWindowRect(hwnd, byref(rect)) == 0:  # NOSONAR
                log_fn("GetWindowRect failure")
                return True
            left, top, right, bottom = int(rect.left), int(rect.top), int(rect.right), int(rect.bottom)
            if right < 0 or bottom < 0:
                log_fn("skipped offscreen window at %ix%i", right, bottom)
                return True
            if hwnd == taskbar:
                log_fn("skipped taskbar")
                return True
            # dirty way:
            if window_title == 'Program Manager':
                return True
            # this should be the proper way using GetTitleBarInfo (but does not seem to work)
            # import ctypes
            # from ctypes.windll.user32 import GetTitleBarInfo
            # from ctypes.wintypes import (DWORD, RECT)
            # class TITLEBARINFO(ctypes.Structure):
            #    pass
            # TITLEBARINFO._fields_ = [
            #    ('cbSize', DWORD),
            #    ('rcTitleBar', RECT),
            #    ('rgstate', DWORD * 6),
            # ]
            # ti = TITLEBARINFO()
            # ti.cbSize = sizeof(ti)
            # GetTitleBarInfo(hwnd, byref(ti))
            # if ti.rgstate[0] & win32con.STATE_SYSTEM_INVISIBLE:
            #    log("skipped system invisible window")
            #    return True
            w = right - left
            h = bottom - top
            log_fn("shape(%s - %#x)=%s", window_title, hwnd, (left, top, w, h))
            if w <= 0 and h <= 0:
                log_fn("skipped invalid window size: %ix%i", w, h)
                return True
            if left == -32000 and top == -32000:
                # there must be a better way of skipping those - I haven't found it
                log_fn("skipped special window")
                return True
            # now clip rectangle:
            if left < 0:
                left = 0
                w = right
            if top < 0:
                top = 0
                h = bottom
            rectangles.append((left, top, w, h))
            return True

        EnumWindows(EnumWindowsProc(enum_windows_cb), 0)
        log_fn("get_shape_rectangles()=%s", rectangles)
        return sorted(rectangles)

    def get_property(self, prop: str):
        if prop == "shape":
            shape = {"Bounding.rectangles": self.rectangles}
            # provide clip rectangle? (based on workspace area?)
            return shape
        return super().get_property(prop)


class Win32ShadowModel(RootWindowModel):
    __slots__ = ("hwnd", "iconic")

    def __init__(self, root_window, capture=None, title="", geometry=None):
        super().__init__(root_window, capture, title, geometry)
        self.hwnd = 0
        self.iconic = geometry[2] == -32000 and geometry[3] == -32000
        self.property_names.append("hwnd")
        self.dynamic_property_names.append("size-hints")

    def get_id(self) -> int:
        return self.hwnd

    def __repr__(self):
        return "Win32ShadowModel(%s : %24s : %s)" % (self.capture, self.geometry, self.hwnd)


class ShadowServer(GTKShadowServerBase):

    def __init__(self, display, attrs: dict[str, str]):
        super().__init__(attrs)
        self.pixel_depth = 32
        self.cursor_handle = None
        self.cursor_data = None
        self.cursor_errors = [0.0, 0]
        self.backend = attrs.get("backend", "auto")
        if GetSystemMetrics(win32con.SM_SAMEDISPLAYFORMAT) == 0:
            raise InitException("all the monitors must use the same display format")
        el = get_win32_event_listener()
        # TODO: deal with those messages?
        el.add_event_callback(win32con.WM_POWERBROADCAST, self.power_broadcast_event)
        # el.add_event_callback(WM_WTSSESSION_CHANGE,         self.session_change_event)
        # these are bound to callbacks in the client,
        # but on the server we just ignore them:
        el.ignore_events.update({
            win32con.WM_ACTIVATEAPP: "WM_ACTIVATEAPP",
            win32con.WM_MOVE: "WM_MOVE",
            win32con.WM_INPUTLANGCHANGE: "WM_INPUTLANGCHANGE",
            win32con.WM_WININICHANGE: "WM_WININICHANGE",
        })

    def init(self, opts) -> None:
        self.pixel_depth = int(opts.pixel_depth) or 32
        if self.pixel_depth not in (24, 30, 32):
            raise InitException("unsupported pixel depth: %s" % self.pixel_depth)
        super().init(opts)

    def power_broadcast_event(self, wParam: int, lParam: int) -> None:
        log("WM_POWERBROADCAST: %s/%s", POWER_EVENTS.get(wParam, wParam), lParam)
        if wParam == win32con.PBT_APMSUSPEND:
            log.info("WM_POWERBROADCAST: PBT_APMSUSPEND")
            for source in self._server_sources.values():
                source.may_notify(NotificationID.IDLE, "Server Suspending",
                                  "This Xpra server is going to suspend,\nthe connection is likely to be interrupted soon.",
                                  expire_timeout=10 * 1000, icon_name="shutdown")
        elif wParam == win32con.PBT_APMRESUMEAUTOMATIC:
            log.info("WM_POWERBROADCAST: PBT_APMRESUMEAUTOMATIC")

    def guess_session_name(self, _procs=None) -> None:
        desktop_name = get_desktop_name()
        log("get_desktop_name()=%s", desktop_name)
        if desktop_name:
            self.session_name = desktop_name

    def print_screen_info(self) -> None:
        size = self.get_root_window_size()
        if not size:
            # we probably don't have access to the screen
            return
        w, h = size
        try:
            display = prettify_plug_name(self.root.get_screen().get_display().get_name())
        except Exception:
            display = ""
        self.do_print_screen_info(display, w, h)

    def make_tray_widget(self):
        from xpra.platform.win32.tray import Win32Tray
        return Win32Tray(self, XPRA_APP_ID, self.tray_menu, "Xpra Shadow Server", "server-notconnected",
                         click_cb=self.tray_click_callback, exit_cb=self.tray_exit_callback)

    def setup_capture(self):
        w, h = get_root_window_size()
        capture = try_setup_capture(CAPTURE_BACKENDS, self.backend, w, h, self.pixel_depth)
        log(f"setup_capture() {self.backend} - {self.root}: {capture}")
        return capture

    def get_root_window_model_class(self) -> type:
        if SEAMLESS:
            return SeamlessRootWindowModel
        return RootWindowModel

    def makeDynamicWindowModels(self):
        assert self.window_matches
        ourpid = os.getpid()
        taskbar = FindWindowA("Shell_TrayWnd", None)
        windows: dict[int, tuple[str, tuple[int, int, int, int]]] = {}

        def enum_windows_cb(hwnd, lparam):
            if not IsWindowVisible(hwnd):
                log("window %#x is not visible", hwnd)
                return True
            pid = c_ulong()
            thread_id = GetWindowThreadProcessId(hwnd, byref(pid))
            if pid == ourpid:
                log("skipped our own window %#x, thread id=%#x", hwnd, thread_id)
                return True
            rect = RECT()
            if GetWindowRect(hwnd, byref(rect)) == 0:  # NOSONAR
                log("GetWindowRect failure")
                return True
            if hwnd == taskbar:
                log("skipped taskbar")
                return True
            # skipping IsWindowEnabled check
            length = GetWindowTextLengthW(hwnd)
            buf = create_unicode_buffer(length + 1)
            window_title = ''
            if GetWindowTextW(hwnd, buf, length + 1) > 0:
                window_title = buf.value
            left, top, right, bottom = int(rect.left), int(rect.top), int(rect.right), int(rect.bottom)
            w = right - left
            h = bottom - top
            if left <= -32000 or top <= -32000:
                log("%r is not visible: %s", window_title, (left, top, w, h))
            if w <= 0 and h <= 0:
                log("skipped invalid window size: %ix%i", w, h)
                return True
            windows[hwnd] = (window_title, (left, top, w, h))
            return True

        EnumWindows(EnumWindowsProc(enum_windows_cb), 0)
        log("makeDynamicWindowModels() windows=%s", windows)
        models = []

        def add_model(hwnd: int, title: str, geometry: tuple[int, int, int, int]):
            model = Win32ShadowModel(self.root, self.capture, title=title, geometry=geometry)
            model.hwnd = hwnd
            models.append(model)

        for m in self.window_matches:
            window = None
            try:
                if m.startswith("0x"):
                    hwnd = int(m, 16)
                else:
                    hwnd = int(m)
                if hwnd:
                    window = windows.pop(hwnd, None)
                    if window:
                        add_model(hwnd, *window)
            except ValueError:
                namere = re.compile(m, re.IGNORECASE)
                for hwnd, window in tuple(windows.items()):
                    title, geometry = window
                    if namere.match(title):
                        add_model(hwnd, title, geometry)
                        windows.pop(hwnd)
        log("makeDynamicWindowModels()=%s", models)
        return models

    def get_shadow_monitors(self) -> list:
        # convert to the format expected by GTKShadowServerBase:
        monitors = []
        for i, monitor in enumerate(get_monitors()):
            geom = monitor["Monitor"]
            x1, y1, x2, y2 = geom
            assert x1 < x2 and y1 < y2
            plug_name = monitor["Device"].lstrip("\\\\.\\")
            monitors.append((plug_name, x1, y1, x2 - x1, y2 - y1, 1))
            screenlog("monitor %i: %10s coordinates: %s", i, plug_name, geom)
        log("get_shadow_monitors()=%s", monitors)
        return monitors

    def refresh(self) -> bool:
        v = super().refresh()
        if v and SEAMLESS:
            for rwm in self._id_to_window.values():
                rwm.refresh_shape()
        log("refresh()=%s", v)
        return v

    def do_get_cursor_data(self) -> tuple | None:
        ci = CURSORINFO()
        ci.cbSize = sizeof(CURSORINFO)
        GetCursorInfo(byref(ci))
        # cursorlog("GetCursorInfo handle=%#x, last handle=%#x", ci.hCursor or 0, self.cursor_handle or 0)
        if not (ci.flags & win32con.CURSOR_SHOWING):
            # cursorlog("do_get_cursor_data() cursor not shown")
            return None
        handle = int(ci.hCursor)
        if handle == self.cursor_handle and self.last_cursor_data:
            # cursorlog("do_get_cursor_data() cursor handle unchanged")
            return self.last_cursor_data
        self.cursor_handle = handle
        cd = get_cursor_data(handle)
        if not cd:
            cursorlog("do_get_cursor_data() no cursor data")
            return self.last_cursor_data
        cd[0] = ci.ptScreenPos.x
        cd[1] = ci.ptScreenPos.y
        w, h = get_fixed_cursor_size()
        return (
            cd,
            ((w, h), [(w, h), ]),
        )

    def get_pointer_position(self) -> tuple[int, int]:
        pos = POINT()
        GetPhysicalCursorPos(byref(pos))  # NOSONAR
        return pos.x, pos.y

    def do_process_mouse_common(self, proto, device_id, wid: int, pointer, props) -> bool:
        ss = self._server_sources.get(proto)
        if not ss:
            return False
        # adjust pointer position for offset in client:
        try:
            x, y = pointer[:2]
            if SetPhysicalCursorPos(x, y):
                return True
            # rate limit the warnings:
            start, count = self.cursor_errors
            now = monotonic()
            elapsed = now - start
            if count == 0 or (count > 1 and elapsed > 10):
                log.warn("Warning: cannot move cursor")
                log.warn(" (%i events)", count + 1)
                self.cursor_errors = [now, 1]
            else:
                self.cursor_errors[1] = count + 1
        except Exception as e:
            log("SetPhysicalCursorPos%s failed", pointer, exc_info=True)
            log.error("Error: failed to move the cursor:")
            log.estr(e)
        return False

    def clear_keys_pressed(self) -> None:
        # noinspection PyCallingNonCallable,PyTypeChecker
        keystate = (BYTE * 256)()
        if GetKeyboardState(keystate):
            vknames = {}
            for vkconst in (x for x in dir(win32con) if x.startswith("VK_")):
                vknames[getattr(win32con, vkconst)] = vkconst[3:]
            pressed = []
            for i in range(256):
                if keystate[i]:
                    pressed.append(vknames.get(i, i))
            keylog("keys still pressed: %s", csv(pressed))
            for x in (
                    win32con.VK_LSHIFT, win32con.VK_RSHIFT, win32con.VK_SHIFT,
                    win32con.VK_LCONTROL, win32con.VK_RCONTROL, win32con.VK_CONTROL,
                    win32con.VK_LMENU, win32con.VK_RMENU, win32con.VK_MENU,
                    win32con.VK_LWIN, win32con.VK_RWIN,
            ):
                keystate[x] = 0
            SetKeyboardState(keystate)

    def get_keyboard_config(self, _props=None) -> KeyboardConfig:
        return KeyboardConfig()

    def fake_key(self, keycode: int, press: bool) -> None:
        fake_key(keycode, press)

    def do_process_button_action(self, proto, device_id, wid: int, button: int, pressed: bool, pointer, props) -> None:
        if "modifiers" in props:
            self._update_modifiers(proto, wid, props.get("modifiers"))
        # ignore device_id on win32:
        did = -1
        pointer = self.process_mouse_common(proto, did, wid, pointer)
        if pointer:
            self.get_server_source(proto).user_event()
            self.button_action(did, wid, pointer, button, pressed, props)

    def button_action(self, device_id, wid: int, pointer, button: int, pressed: bool, props) -> None:
        event = BUTTON_EVENTS.get((button, pressed))
        if event is None:
            log.warn("no matching event found for button=%s, pressed=%s", button, pressed)
            return
        elif event == NOEVENT:
            return
        dwFlags, dwData = event
        x, y = pointer[:2]
        mouse_event(dwFlags, x, y, dwData, 0)

    def make_hello(self, source) -> dict[str, Any]:
        capabilities = GTKServerBase.make_hello(self, source)
        capabilities["shadow"] = True
        capabilities["server_type"] = "Python/Win32-Shadow"
        return capabilities

    def get_info(self, proto, *_args) -> dict[str, Any]:
        info = GTKServerBase.get_info(self, proto)
        info.update(GTKShadowServerBase.get_info(self, proto))
        info.setdefault("features", {})["shadow"] = True
        info.setdefault("server", {
            "pixel-depth": self.pixel_depth,
            "type": "Python/Win32-Shadow",
            "tray": self.tray,
            "tray-icon": self.tray_icon or ""
        })
        return info


def main():
    from xpra.platform import program_context
    with program_context("Shadow-Test", "Shadow Server Screen Capture Test"):
        rwm = RootWindowModel(None)
        pngdata = rwm.take_screenshot()
        filename = "screenshot.png"
        with open(filename, "wb") as f:
            f.write(pngdata[4])
        print(f"saved screenshot as {filename}")


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2018 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

"""
Utility functions for loading icons from files
"""

import os
import re
from io import BytesIO

from xpra.util.str_fn import Ellipsizer
from xpra.util.env import envint, envbool, first_time, SilenceWarningsContext
from xpra.os_util import gi_import
from xpra.util.io import load_binary_file
from xpra.log import Logger

log = Logger("menu")

MAX_ICON_SIZE = envint("XPRA_XDG_MAX_ICON_SIZE", 0)
SVG_TO_PNG = envbool("XPRA_SVG_TO_PNG", True)

INKSCAPE_RE = b'\\sinkscape:[a-zA-Z]*=["a-zA-Z0-9]*'
INKSCAPE_BROKEN_SODIPODI_DTD = b'xmlns:sodipodi="http://inkscape.sourceforge.net/DTD/s odipodi-0.dtd"'
INKSCAPE_SODIPODI_DTD = b'xmlns:sodipodi="http://inkscape.sourceforge.net/DTD/sodipodi-0.dtd"'

large_icons = []

_rsvg = None


def load_rsvg():
    global _rsvg
    if _rsvg is None:
        try:
            rsvg = gi_import("Rsvg")
            log(f"load_Rsvg() {rsvg=}")
            _rsvg = rsvg
        except (ValueError, ImportError) as e:
            if first_time("no-rsvg"):
                log.warn("Warning: cannot resize svg icons,")
                log.warn(" the Rsvg bindings were not found:")
                log.warn(" %s", e)
            _rsvg = False
    return _rsvg


def load_icon_from_file(filename: str, max_size: int = MAX_ICON_SIZE) -> tuple:
    if os.path.isdir(filename):
        log("load_icon_from_file(%s, %i) path is a directory!", filename, max_size)
        return ()
    log("load_icon_from_file(%s, %i)", filename, max_size)
    if filename.endswith("xpm"):
        img = None
        try:
            from PIL import Image  # pylint: disable=import-outside-toplevel
            img = Image.open(filename)
            buf = BytesIO()
            img.save(buf, "PNG")
            pngicondata = buf.getvalue()
            buf.close()
            return pngicondata, "png"
        except (ValueError, ImportError) as e:
            log(f"Image.open({filename}) {e}", exc_info=True)
        except Exception as e:
            log(f"Image.open({filename})", exc_info=True)
            log.error(f"Error loading {filename!r}:")
            log.estr(e)
        finally:
            if img:
                img.close()
    icondata = load_binary_file(filename)
    if not icondata:
        return ()
    if filename.endswith("svg") and max_size and len(icondata) > max_size:
        # try to resize it
        size = len(icondata)
        pngdata = svg_to_png(filename, icondata)
        if pngdata:
            log(f"reduced size of SVG icon {filename}, from {size} bytes to {len(pngdata)} bytes as PNG")
            icondata = pngdata
            filename = filename[:-3] + "png"
    log("got icon data from '%s': %i bytes", filename, len(icondata))
    if 0 < max_size < len(icondata) and first_time(f"icon-size-warning-{filename}"):
        large_icons.append((filename, len(icondata)))
    return icondata, os.path.splitext(filename)[1].lstrip(".")


def svg_to_png(filename: str, icondata, w: int = 128, h: int = 128) -> bytes:
    if not SVG_TO_PNG:
        return b""
    rsvg = load_rsvg()
    log("svg_to_png%s rsvg=%s", (filename, f"{len(icondata)} bytes", w, h), rsvg)
    if not rsvg:
        return b""
    try:
        # pylint: disable=no-name-in-module, import-outside-toplevel
        from cairo import ImageSurface, Context, Format
        img = ImageSurface(Format.ARGB32, w, h)
        ctx = Context(img)
        handle = rsvg.Handle.new_from_data(data=icondata)
        if hasattr(handle, "render_document"):
            # Rsvg version 2.46 and later:
            rect = rsvg.Rectangle()
            rect.x = rect.y = 0
            rect.width = w
            rect.height = h
            if not handle.render_document(ctx, rect):
                raise RuntimeError(f"{handle}.render_document failed")
        else:
            with SilenceWarningsContext():
                dim = handle.get_dimensions()
                ctx.scale(w / dim.width, h / dim.height)
                handle.render_cairo(ctx)
        del handle
        img.flush()
        buf = BytesIO()
        img.write_to_png(buf)
        icondata = buf.getvalue()
        buf.close()
        img.finish()
        return icondata
    except Exception:
        log("svg_to_png%s", (icondata, w, h), exc_info=True)
        if re.findall(INKSCAPE_RE, icondata):
            # try again after stripping the bogus inkscape attributes
            # as some rsvg versions can't handle that (ie: Debian Bullseye)
            icondata = re.sub(INKSCAPE_RE, b"", icondata)
            return svg_to_png(filename, icondata, w, h)
        if icondata.find(INKSCAPE_BROKEN_SODIPODI_DTD) > 0:
            icondata = icondata.replace(INKSCAPE_BROKEN_SODIPODI_DTD, INKSCAPE_SODIPODI_DTD)
            return svg_to_png(filename, icondata, w, h)
        log.error("Error: failed to convert svg icon")
        if filename:
            log.error(" '%s':", filename)
        log.error(" %i bytes, %s", len(icondata), Ellipsizer(icondata))
        return b""



# This file is part of Xpra.
# Copyright (C) 2020 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from xpra.os_util import gi_import
from xpra.net.qrcode import qrencode
from xpra.gtk.window import add_close_accel
from xpra.log import Logger, consume_verbose_argv

Gtk = gi_import("Gtk")
Gdk = gi_import("Gdk")
GLib = gi_import("GLib")
GdkPixbuf = gi_import("GdkPixbuf")

log = Logger("menu")


def show_qr(uri: str, width: int = 640, height: int = 640):
    assert uri.find(":") > 0, "invalid uri"
    # support old-style URIs, ie: tcp:host:port
    if uri.find(":") != uri.find("://"):
        uri = uri.replace(":", "://", 1)
    parts = uri.split(":", 1)
    if parts[0] in ("tcp", "ws"):
        uri = "http:" + parts[1]
    else:
        uri = "https:" + parts[1]
    pixbuf = qr_pixbuf(uri, width, height)
    if not pixbuf:
        return
    image = Gtk.Image().new_from_pixbuf(pixbuf)
    window = Gtk.Window(modal=True, title="QR Code")
    window.set_position(Gtk.WindowPosition.CENTER)
    window.add(image)
    window.set_size_request(width, height)
    window.set_resizable(False)

    def close(*_args):
        window.close()

    add_close_accel(window, close)
    window.show_all()
    if Gtk.main_level() <= 0:
        def gtk_quit(*_args):
            close()
            Gtk.main_quit()

        window.connect("delete-event", gtk_quit)
        Gtk.main()


def qr_pixbuf(uri: str, width: int = 640, height: int = 640):
    img = qrencode.encode_image(uri)
    if not img:
        return None
    from PIL import Image
    try:
        NEAREST = Image.Resampling.NEAREST
    except AttributeError:
        NEAREST = Image.NEAREST
    img = img.convert("RGB")
    img = img.resize((width, height), NEAREST)
    data = img.tobytes()
    w, h = img.size
    data = GLib.Bytes.new(data)
    pixbuf = GdkPixbuf.Pixbuf.new_from_bytes(data, GdkPixbuf.Colorspace.RGB,
                                             False, 8, w, h, w * 3)
    return pixbuf


def main(args: list[str]) -> int:
    consume_verbose_argv(args, "menu")
    uri = sys.argv[1]
    show_qr(uri)
    return 0


if __name__ == "__main__":
    import sys
    sys.exit(main(sys.argv))



# This file is part of Xpra.
# Copyright (C) 2018 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import re
import socket
import sys
from time import sleep, monotonic
from typing import Any, NoReturn
from collections.abc import Sequence

from xpra.scripts.main import InitException, InitExit, shellquote, host_target_string
from xpra.platform.paths import get_ssh_known_hosts_files
from xpra.platform.info import get_username
from xpra.scripts.config import str_to_bool, TRUE_OPTIONS
from xpra.scripts.pinentry import input_pass, confirm
from xpra.net.ssh.util import get_default_keyfiles, LOG_EOF
from xpra.net.bytestreams import SocketConnection, SOCKET_TIMEOUT
from xpra.util.thread import start_thread
from xpra.exit_codes import ExitCode
from xpra.util.io import load_binary_file, stderr_print, umask_context
from xpra.common import noerr
from xpra.util.str_fn import csv
from xpra.util.env import envint, envbool, envfloat, first_time
from xpra.log import Logger

# pylint: disable=import-outside-toplevel

log = Logger("network", "ssh")
if log.is_debug_enabled():
    import logging

    logging.getLogger("paramiko").setLevel(logging.DEBUG)

from paramiko.ssh_exception import SSHException, ProxyCommandFailure, PasswordRequiredException  # noqa: E402
from paramiko.transport import Transport  # noqa: E402

WINDOW_SIZE = envint("XPRA_SSH_WINDOW_SIZE", 2 ** 27 - 1)
TIMEOUT = envint("XPRA_SSH_TIMEOUT", 60)

VERIFY_HOSTKEY = envbool("XPRA_SSH_VERIFY_HOSTKEY", True)
VERIFY_STRICT = envbool("XPRA_SSH_VERIFY_STRICT", False)
ADD_KEY = envbool("XPRA_SSH_ADD_KEY", True)
# which authentication mechanisms are enabled with paramiko:
NONE_AUTH = envbool("XPRA_SSH_NONE_AUTH", True)
PASSWORD_AUTH = envbool("XPRA_SSH_PASSWORD_AUTH", True)
AGENT_AUTH = envbool("XPRA_SSH_AGENT_AUTH", True)
KEY_AUTH = envbool("XPRA_SSH_KEY_AUTH", True)
PASSWORD_RETRY = envint("XPRA_SSH_PASSWORD_RETRY", 2)
SSH_AGENT = envbool("XPRA_SSH_AGENT", False)
BANNER = envbool("XPRA_SSH_BANNER", True)
assert PASSWORD_RETRY >= 0
LOG_FAILED_CREDENTIALS = envbool("XPRA_LOG_FAILED_CREDENTIALS", False)
TEST_COMMAND_TIMEOUT = envint("XPRA_SSH_TEST_COMMAND_TIMEOUT", 10)
EXEC_STDOUT_TIMEOUT = envfloat("XPRA_SSH_EXEC_STDOUT_TIMEOUT", 2)
EXEC_STDERR_TIMEOUT = envfloat("XPRA_SSH_EXEC_STDERR_TIMEOUT", 0)

MSYS_DEFAULT_PATH = os.environ.get("XPRA_MSYS_DEFAULT_PATH", "/mingw64/bin/xpra")
CYGWIN_DEFAULT_PATH = os.environ.get("XPRA_CYGWIN_DEFAULT_PATH", "/cygdrive/c/Program Files/Xpra/Xpra_cmd.exe")
DEFAULT_WIN32_INSTALL_PATH = "C:\\Program Files\\Xpra"
WIN_OPENSSH_AGENT = envbool("XPRA_WIN_OPENSSH_AGENT", False)
WIN_OPENSSH_AGENT_MODULE = "win_openssh.OpenSSHAgentConnection"

PARAMIKO_SESSION_LOST = "No existing session"

WIN32_REGISTRY_QUERY = "REG QUERY \"HKEY_LOCAL_MACHINE\\Software\\Xpra\" /v InstallPath"

AUTH_MODES: Sequence[str] = ("none", "agent", "key", "password")


def keymd5(k) -> str:
    import binascii
    f = binascii.hexlify(k.get_fingerprint()).decode("latin1")
    s = "MD5"
    while f:
        s += ":" + f[:2]
        f = f[2:]
    return s


if BANNER:
    from paramiko import auth_handler

    class AuthHandlerOverride(auth_handler.AuthHandler):

        def _parse_userauth_banner(self, m):
            banner = m.get_string()
            self.banner = banner
            if banner:
                try:
                    text = banner.decode("utf8")
                except UnicodeDecodeError:
                    text = repr(banner)
                log.info("SSH Authentication Banner:")
                for msg in text.splitlines():
                    log.info(f" {msg!r}")

    auth_handler.AuthHandler = AuthHandlerOverride
    from paramiko import transport
    transport.AuthHandler = AuthHandlerOverride


class SSHSocketConnection(SocketConnection):

    def __init__(self, ssh_channel, sock, sockname: str | tuple | list,
                 peername, target, info=None, socket_options=None):
        self._raw_socket = sock
        super().__init__(ssh_channel, sockname, peername, target, "ssh", info, socket_options)

    def get_raw_socket(self):
        return self._raw_socket

    def start_stderr_reader(self) -> None:
        start_thread(self._stderr_reader, "ssh-stderr-reader", daemon=True)

    def _stderr_reader(self) -> None:
        chan = self._socket
        stderr = chan.makefile_stderr("rb", 1)
        while self.active:
            v = stderr.readline()
            if not v:
                if LOG_EOF and self.active:
                    log.info("SSH EOF on stderr of %s", chan.get_name())
                break
            s = v.rstrip(b"\n\r").decode()
            if s:
                log.info(" SSH: %r", s)

    def peek(self, n) -> bytes:
        if not self._raw_socket:
            return b""
        return self._raw_socket.recv(n, socket.MSG_PEEK)

    def get_socket_info(self) -> dict[str, Any]:
        if not self._raw_socket:
            return {}
        return self.do_get_socket_info(self._raw_socket)

    def get_info(self) -> dict[str, Any]:
        i = super().get_info()
        s = self._socket
        if s:
            i["ssh-channel"] = {
                "id": s.get_id(),
                "name": s.get_name(),
            }
        return i


class SSHProxyCommandConnection(SSHSocketConnection):
    def __init__(self, ssh_channel, peername, target, info):
        super().__init__(ssh_channel, None, "", peername, target, info)
        self.process = None

    def error_is_closed(self, e) -> bool:
        p = self.process
        if p:
            # if the process has terminated,
            # then the connection must be closed:
            if p[0].poll() is not None:
                return True
        return super().error_is_closed(e)

    def get_socket_info(self) -> dict[str, Any]:
        p = self.process
        if not p:
            return {}
        proc, _ssh, cmd = p
        return {
            "process": {
                "pid": proc.pid,
                "returncode": proc.returncode,
                "command": cmd,
            }
        }

    def close(self) -> None:
        try:
            super().close()
        except OSError:
            # this can happen if the proxy command gets a SIGINT,
            # it's closed already and we don't care
            log("SSHProxyCommandConnection.close()", exc_info=True)


def safe_lookup(config_obj, hostname: str) -> dict:
    try:
        _lookup = getattr(config_obj, "_lookup", None)
        import paramiko
        # older versions don't have the same signature for `_lookup`:
        if _lookup and getattr(paramiko, "__version_info__", (0, )) >= (3, 3):
            # completely duplicate the paramiko logic since they're unwilling to merge a trivial change :(
            options = _lookup(hostname=hostname)
            # Inject HostName if it was not set (this used to be done incidentally
            # during tokenization, for some reason).
            if "hostname" not in options:
                options["hostname"] = hostname
            # Handle canonicalization
            canon = options.get("canonicalizehostname", None) in ("yes", "always")
            maxdots = int(options.get("canonicalizemaxdots", 1))
            if canon and hostname.count(".") <= maxdots:
                # NOTE: OpenSSH manpage does not explicitly state this, but its
                # implementation for CanonicalDomains is 'split on any whitespace'.
                domains = options.get("canonicaldomains", "").split()
                hostname = config_obj.canonicalize(hostname, options, domains)
                # Overwrite HostName again here (this is also what OpenSSH does)
                options["hostname"] = hostname
                options = _lookup(
                    hostname, options, canonical=True, final=True
                )
            else:
                options = _lookup(
                    hostname, options, canonical=False, final=True
                )
        else:
            options = config_obj.lookup(hostname)
        return dict(options or {})
    except ImportError as e:
        log("%s.lookup(%s)", config_obj, hostname, exc_info=True)
        log.warn(f"Warning: unable to load SSH host config for {hostname!r}:")
        log.warn(f" {e}")
        if isinstance(e, ModuleNotFoundError):
            log.warn(" (looks like a 'paramiko' distribution packaging issue)")
    except KeyError as e:
        log("%s.lookup(%s)", config_obj, hostname, exc_info=True)
        log.info(f"paramiko ssh config lookup error for host {hostname!r}")
        if first_time("paramiko-#2338"):
            log.info(" %s: %s", type(e), e)
            log.info(" the paramiko project looks unmaintained:")
            log.info(" https://github.com/paramiko/paramiko/pull/2338")
    return {}


def connect_to(display_desc: dict) -> SSHSocketConnection:
    log(f"connect_to({display_desc})")
    # plain socket attributes:
    host: str = display_desc["host"]
    port: int = display_desc.get("port", 22)
    peername: tuple[str, int] | tuple[str, int, int, int] = (host, port)
    # ssh and command attributes:
    username: str = display_desc.get("username") or get_username()
    if "proxy_host" in display_desc:
        display_desc.setdefault("proxy_username", get_username())
    password: str = display_desc.get("password", "")
    remote_xpra = display_desc["remote_xpra"]
    proxy_command = display_desc["proxy_command"]  # ie: "_proxy_start"
    socket_dir: str = display_desc.get("socket_dir", "")
    display: str = display_desc.get("display", "")
    display_as_args: list[str] = display_desc["display_as_args"]  # ie: "--start=xterm :10"
    paramiko_config: dict = display_desc.copy()
    paramiko_config.update(display_desc.get("paramiko-config", {}))
    socket_info: dict[str, Any] = {
        "host": host,
        "port": port,
    }

    def fail(msg: str) -> NoReturn:
        log("connect_to(%s)", display_desc, exc_info=True)
        raise InitExit(ExitCode.SSH_FAILURE, msg) from None

    from paramiko import SSHConfig, ProxyCommand
    ssh_config = SSHConfig()

    etc = "etc" if sys.prefix == "/usr" else sys.prefix + "/etc"
    for config_file in (
            f"{etc}/ssh/ssh_config",
            os.path.expanduser("~/.ssh/config"),
    ):
        if not os.path.exists(config_file):
            continue
        with open(config_file, encoding="utf8") as f:
            try:
                ssh_config.parse(f)
            except Exception as e:
                log(f"parse({config_file})", exc_info=True)
                log.error(f"Error parsing {config_file!r}:")
                log.estr(e)
        log(f"parsed ssh config {config_file!r}")
    try:
        log("%i hosts found", len(ssh_config.get_hostnames()))
    except KeyError:
        pass

    def ssh_lookup(key) -> dict:
        return safe_lookup(ssh_config, key)

    host_config: dict = ssh_lookup("*")
    host_config.update(ssh_lookup(host))
    log(f"host_config({host})={host_config}")

    def get_keyfiles(config_name="key") -> list[str]:
        keyfiles = host_config.get("identityfile") or get_default_keyfiles()
        keyfile = paramiko_config.get(config_name)
        if keyfile:
            keyfiles.insert(0, keyfile)
        return keyfiles

    if host_config:
        # paramiko does not strip comments from configs:
        # https://github.com/Xpra-org/xpra/issues/4503
        def safeget(key: str, default_value: str | int) -> Any:
            value = host_config.get(key, default_value)
            if isinstance(value, str):
                value = value.split("#", 1)[0]
            return value
        log(f"got host config for {host!r}: {host_config}")
        host = safeget("hostname", host)
        username = safeget("user", username)
        port = safeget("port", port)
        try:
            port = int(port)
        except (TypeError, ValueError):
            raise InitExit(ExitCode.SSH_FAILURE, f"invalid ssh port specified: {port!r}") from None
        proxycommand = host_config.get("proxycommand")
        if proxycommand:
            log(f"found proxycommand={proxycommand!r} for host {host!r}")
            sock = ProxyCommand(proxycommand)
            log(f"ProxyCommand({proxycommand})={sock}")
            from xpra.util.child_reaper import getChildReaper
            cmd = getattr(sock, "cmd", [])

            def proxycommand_ended(proc):
                log(f"proxycommand_ended({proc}) exit code={proc.poll()}")

            getChildReaper().add_process(sock.process, "paramiko-ssh-client", cmd, True, True,
                                         callback=proxycommand_ended)
            proxy_keys = get_keyfiles("proxy_key")
            log(f"{proxy_keys=}")
            from paramiko.client import SSHClient
            ssh_client = SSHClient()
            ssh_client.load_system_host_keys()
            log("ssh proxy command connect to %s", (host, port, sock))
            ssh_client.connect(host, port, sock=sock)
            transport = ssh_client.get_transport()
            do_connect_to(transport, host,
                          username, password,
                          host_config,
                          proxy_keys,
                          paramiko_config)
            chan = run_remote_xpra(transport, proxy_command, remote_xpra, socket_dir, display_as_args)
            peername = (host, port)
            conn = SSHProxyCommandConnection(chan, peername, peername, socket_info)
            conn.target = host_target_string("ssh", username, host, port, display)
            conn.timeout = SOCKET_TIMEOUT
            conn.start_stderr_reader()
            conn.process = (sock.process, "ssh", cmd)
            from xpra.net import bytestreams
            bytestreams.CLOSED_EXCEPTIONS = tuple(list(bytestreams.CLOSED_EXCEPTIONS) + [ProxyCommandFailure])
            return conn

    keys = get_keyfiles()
    from xpra.net.socket_util import socket_connect
    if "proxy_host" in display_desc:
        proxy_host = display_desc["proxy_host"]
        proxy_port = int(display_desc.get("proxy_port", 22))
        proxy_username = display_desc.get("proxy_username", username)
        proxy_password = display_desc.get("proxy_password", password)
        proxy_keys = get_keyfiles("proxy_key")
        sock = socket_connect(proxy_host, proxy_port)
        if not sock:
            fail(f"SSH proxy transport failed to connect to {proxy_host}:{proxy_port}")
        middle_transport = do_connect(sock, proxy_host,
                                      proxy_username, proxy_password,
                                      ssh_lookup(host) or ssh_lookup("*"),
                                      proxy_keys,
                                      paramiko_config)
        log("Opening proxy channel")
        chan_to_middle = middle_transport.open_channel("direct-tcpip", (host, port), ("localhost", 0))
        transport = do_connect(chan_to_middle, host,
                               username, password,
                               host_config,
                               keys,
                               paramiko_config)
        chan = run_remote_xpra(transport, proxy_command, remote_xpra, socket_dir, display_as_args)
        conn = SSHProxyCommandConnection(chan, peername, peername, socket_info)
        to_str = host_target_string("ssh", username, host, port, display)
        proxy_str = host_target_string("ssh", proxy_username, proxy_host, proxy_port)
        conn.target = f"{to_str} via {proxy_str}"
        conn.timeout = SOCKET_TIMEOUT
        conn.start_stderr_reader()
        return conn

    # plain TCP connection to the server,
    # we open it then give the socket to paramiko:
    auth_modes = get_auth_modes(paramiko_config, host_config, password)
    log(f"authentication modes={auth_modes}")
    sock = None
    transport = None
    sockname = host
    while not transport:
        sock = socket_connect(host, port)
        if not sock:
            fail(f"SSH failed to connect to {host}:{port}")
        sockname = sock.getsockname()
        peername = sock.getpeername()
        log(f"paramiko socket_connect: sockname={sockname}, peername={peername}")
        try:
            transport = do_connect(sock, host, username, password,
                                   host_config,
                                   keys,
                                   paramiko_config,
                                   auth_modes)
        except SSHAuthenticationError as e:
            log(f"paramiko authentication errors on socket {sock} with modes {auth_modes}: {e.errors}",
                exc_info=True)
            pw_errors = []
            for errs in e.errors.values():
                pw_errors += errs
            if ("key" in auth_modes or "agent" in auth_modes) and PARAMIKO_SESSION_LOST in pw_errors:
                # try connecting again but without 'key' and 'agent' authentication:
                # see https://github.com/Xpra-org/xpra/issues/3223
                for m in ("key", "agent"):
                    try:
                        auth_modes.remove(m)
                    except KeyError:
                        pass
                log.info(f"retrying SSH authentication with modes {csv(auth_modes)}")
                continue
            raise
        finally:
            if sock and not transport:
                noerr(sock.shutdown)
                noerr(sock.close)
                sock = None

    remote_port: int = display_desc.get("remote_port", 0)
    if remote_port:
        # we want to connect directly to a remote port,
        # we don't need to run a command
        chan = transport.open_channel("direct-tcpip", ("localhost", remote_port), ('localhost', 0))
        log(f"direct channel to remote port {remote_port} : {chan}")
    else:
        chan = run_remote_xpra(transport, proxy_command, remote_xpra,
                               socket_dir, display_as_args, paramiko_config)
    conn = SSHSocketConnection(chan, sock, sockname, peername, (host, port), socket_info)
    conn.target = host_target_string("ssh", username, host, port, display)
    conn.timeout = SOCKET_TIMEOUT
    conn.start_stderr_reader()
    log(f"paramiko_client.connect_to({display_desc})={conn}")
    return conn


def get_auth_modes(paramiko_config, host_config: dict, password: str) -> list[str]:
    def configvalue(key: str):
        # if the paramiko config has a setting, honour it:
        if paramiko_config and key in paramiko_config:
            return paramiko_config.get(key)
        # fallback to the value from the host config:
        return host_config.get(key)

    def configbool(key: str, default_value=True) -> bool:
        return str_to_bool(configvalue(key), default_value)

    auth_str = configvalue("auth")
    if auth_str:
        return auth_str.split("+")
    auth = []
    if configbool("noneauthentication", NONE_AUTH):
        auth.append("none")
    if password and configbool("passwordauthentication", PASSWORD_AUTH):
        auth.append("password")
    if configbool("agentauthentication", AGENT_AUTH):
        auth.append("agent")
    # Some people do two-factor using KEY_AUTH to kick things off, so this happens first
    if configbool("keyauthentication", KEY_AUTH):
        auth.append("key")
    if not password and configbool("passwordauthentication", PASSWORD_AUTH):
        auth.append("password")
    return auth


class IAuthHandler:
    def __init__(self, password):
        self.authcount = 0
        self.password = password

    def handle_request(self, title: str, instructions, prompt_list) -> list:
        log("handle_request%s counter=%i", (title, instructions, prompt_list), self.authcount)
        p = []
        for pent in prompt_list:
            if self.password:
                p.append(self.password)
                self.password = None
            else:
                p.append(input_pass(pent[0]))
        self.authcount += 1
        log(f"handle_request(..) returning {len(p)} values")
        return p


def do_connect(chan, host: str, username: str, password: str,
               host_config: dict, keyfiles: list[str], paramiko_config: dict, auth_modes=AUTH_MODES):
    transport = Transport(chan)
    transport.use_compression(False)
    log("SSH transport %s", transport)
    try:
        transport.start_client()
    except SSHException as e:
        log("SSH negotiation failed", exc_info=True)
        raise InitExit(ExitCode.SSH_FAILURE, "SSH negotiation failed: %s" % e) from None
    return do_connect_to(transport, host, username, password,
                         host_config, keyfiles, paramiko_config, auth_modes)


def do_connect_to(transport, host: str, username: str, password: str,
                  host_config=None, keyfiles=None, paramiko_config=None, auth_modes=AUTH_MODES):
    log("do_connect_to%s", (transport, host, username, password, host_config, keyfiles, paramiko_config))

    def configvalue(key: str) -> Any:
        # if the paramiko config has a setting, honour it:
        if paramiko_config and key in paramiko_config:
            return paramiko_config.get(key)
        # fallback to the value from the host config:
        return (host_config or {}).get(key)

    def configbool(key: str, default_value=True) -> bool:
        return str_to_bool(configvalue(key), default_value)

    def configint(key: str, default_value=0) -> int:
        v = configvalue(key)
        if v is None:
            return default_value
        return int(v)

    host_key = transport.get_remote_server_key()
    assert host_key, "no remote server key"
    log("remote_server_key=%s", keymd5(host_key))
    if configbool("verify-hostkey", VERIFY_HOSTKEY):
        from paramiko.hostkeys import HostKeys
        host_keys = HostKeys()
        host_keys_filename = None
        known_hosts_files = get_ssh_known_hosts_files()
        for known_hosts in known_hosts_files:
            host_keys.clear()
            try:
                path = os.path.expanduser(known_hosts)
                if os.path.exists(path):
                    host_keys.load(path)
                    log("HostKeys.load(%s) successful", path)
                    host_keys_filename = path
                    break
            except OSError:
                log("HostKeys.load(%s)", known_hosts, exc_info=True)

        log("host keys=%s", host_keys)
        keys = safe_lookup(host_keys, host)
        known_host_key = keys.get(host_key.get_name())

        def keyname():
            return host_key.get_name().replace("ssh-", "")

        if known_host_key and host_key == known_host_key:
            assert host_key
            log("%s host key '%s' OK for host '%s'", keyname(), keymd5(host_key), host)
        else:
            dnscheck = ""
            if configbool("verifyhostkeydns"):
                try:
                    from xpra.net.ssh.sshfp import check_host_key
                    dnscheck = check_host_key(host, host_key)
                except ImportError as e:
                    log("verifyhostkeydns failed", exc_info=True)
                    log.info("cannot check SSHFP DNS records")
                    log.info(" %s", e)
            log("dnscheck=%s", dnscheck)

            def adddnscheckinfo(q: list[str]):
                if dnscheck is not True:
                    if dnscheck:
                        q.append("SSHFP validation failed:")
                        q.append(str(dnscheck))
                    else:
                        q.append("SSHFP validation failed")

            if dnscheck is True:
                # DNSSEC provided a matching record
                log.info("found a valid SSHFP record for host %s", host)
            elif known_host_key:
                log.warn("Warning: SSH server key mismatch")
                qinfo: list[str] = [
                    "WARNING: REMOTE HOST IDENTIFICATION HAS CHANGED!",
                    "IT IS POSSIBLE THAT SOMEONE IS DOING SOMETHING NASTY!",
                    "Someone could be eavesdropping on you right now (man-in-the-middle attack)!",
                    "It is also possible that a host key has just been changed.",
                    f"The fingerprint for the {keyname()} key sent by the remote host is",
                    keymd5(host_key),
                ]
                adddnscheckinfo(qinfo)
                if configbool("stricthostkeychecking", VERIFY_STRICT):
                    log.warn("Host key verification failed.")
                    # TODO: show alert with no option to accept key
                    qinfo += [
                        "Please contact your system administrator.",
                        "Add correct host key in %s to get rid of this message.",
                        f"Offending {keyname()} key in {host_keys_filename}",
                        f"ECDSA host key for {keyname()} has changed and you have requested strict checking.",
                    ]
                    stderr_print(os.linesep.join(qinfo))
                    transport.close()
                    raise InitExit(ExitCode.SSH_KEY_FAILURE, "SSH Host key has changed")
                if not confirm(qinfo):
                    transport.close()
                    raise InitExit(ExitCode.SSH_KEY_FAILURE, "SSH Host key has changed")
                log.info("host key confirmed")
            else:
                assert (not keys) or (host_key.get_name() not in keys)
                if not keys:
                    log.warn("Warning: unknown SSH host '%s'", host)
                else:
                    log.warn("Warning: unknown %s SSH host key", keyname())
                qinfo = [
                    f"The authenticity of host {host!r} can't be established.",
                    f"{keyname()} key fingerprint is",
                    keymd5(host_key),
                ]
                adddnscheckinfo(qinfo)
                if not confirm(qinfo):
                    transport.close()
                    raise InitExit(ExitCode.SSH_KEY_FAILURE, f"Unknown SSH host {host!r}")
                log.info("host key confirmed")
            if configbool("addkey", ADD_KEY) and known_hosts_files:
                try:
                    if not host_keys_filename:
                        # the first one is the default,
                        # ie: ~/.ssh/known_hosts on posix
                        host_keys_filename = os.path.expanduser(known_hosts_files[0])
                    log(f"adding {keyname()} key for host {host!r} to {host_keys_filename!r}")
                    if not os.path.exists(host_keys_filename):
                        keys_dir = os.path.dirname(host_keys_filename)
                        if not os.path.exists(keys_dir):
                            log(f"creating keys directory {keys_dir!r}")
                            os.mkdir(keys_dir, 0o700)
                        elif not os.path.isdir(keys_dir):
                            log.warn(f"Warning: {keys_dir!r} is not a directory")
                            log.warn(" key not saved")
                        if os.path.exists(keys_dir) and os.path.isdir(keys_dir):
                            log(f"creating known host file {host_keys_filename!r}")
                            with umask_context(0o133):
                                with open(host_keys_filename, "ab+"):
                                    "file has been created"
                    host_keys.add(host, host_key.get_name(), host_key)
                    host_keys.save(host_keys_filename)
                except OSError as e:
                    log(f"failed to add key to {host_keys_filename!r}")
                    log.error(f"Error adding key to {host_keys_filename!r}")
                    log.error(f" {e}")
                except Exception:
                    log.error("cannot add key", exc_info=True)
    else:
        log("ssh host key verification skipped")

    auth_errors: dict[str, list[str]] = {}

    def auth_agent() -> None:
        from paramiko.agent import Agent
        agent = Agent()
        agent_keys = agent.get_keys()
        log("agent keys: %s", agent_keys)
        if agent_keys:
            for agent_key in agent_keys:
                log("trying ssh-agent key '%s'", keymd5(agent_key))
                try:
                    transport.auth_publickey(username, agent_key)
                    log("tried ssh-agent key '%s'", keymd5(agent_key))
                    if transport.is_authenticated():
                        log("authenticated using agent and key '%s'", keymd5(agent_key))
                        return
                except AttributeError as e:
                    log(f"auth_publickey({username}, {agent_key})")
                    log.warn("Warning: paramiko bug during public key agent authentication")
                    log.warn(f" {type(e)}: {e}")
                    log.warn(f" using key {type(agent_key)}: {agent_key}")
                except SSHException as e:
                    auth_errors.setdefault("agent", []).append(str(e))
                    log.info("SSH agent key '%s' rejected for user '%s'", keymd5(agent_key), username)
                    log("%s%s", transport.auth_publickey, (username, agent_key), exc_info=True)
                    if str(e) == PARAMIKO_SESSION_LOST:
                        # no point in trying more keys
                        break
            if not transport.is_authenticated():
                log.info("agent authentication failed, tried %i keys", len(agent_keys))

    def auth_publickey() -> None:
        log(f"trying public key authentication using {keyfiles}")
        for keyfile_path in keyfiles:
            if not os.path.exists(keyfile_path):
                log(f"no keyfile at {keyfile_path!r}")
                continue
            log(f"trying {keyfile_path!r}")
            key = None
            import paramiko
            try_key_formats: Sequence[str] = ()
            for kf in ("RSA", "DSS", "ECDSA", "Ed25519"):
                if keyfile_path.lower().endswith(kf.lower()):
                    try_key_formats = (kf,)
                    break
            if not try_key_formats:
                try_key_formats = ("RSA", "DSS", "ECDSA", "Ed25519")
            pkey_classname = None
            for pkey_classname in try_key_formats:
                pkey_class = getattr(paramiko, f"{pkey_classname}Key", None)
                if pkey_class is None:
                    log(f"no {pkey_classname} key type")
                    continue
                log(f"trying to load as {pkey_classname}")
                key = None
                try:
                    key = pkey_class.from_private_key_file(keyfile_path)
                    log.info(f"loaded {pkey_classname} private key from {keyfile_path!r}")
                    break
                except PasswordRequiredException as e:
                    log(f"{keyfile_path!r} keyfile requires a passphrase: {e}")
                    passphrase = input_pass(f"please enter the passphrase for {keyfile_path!r}")
                    if passphrase:
                        try:
                            key = pkey_class.from_private_key_file(keyfile_path, passphrase)
                            log.info(f"loaded {pkey_classname} private key from {keyfile_path!r}")
                        except SSHException as ke:
                            log("from_private_key_file", exc_info=True)
                            log.info(f"cannot load key from file {keyfile_path}:")
                            for emsg in str(ke).split(". "):
                                if emsg.startswith("('"):
                                    emsg = emsg[2:]
                                if emsg.endswith(")."):
                                    emsg = emsg[:-2]
                                if emsg:
                                    log.info(" %s.", emsg)
                    break
                except Exception:
                    log(f"auth_publickey() loading as {pkey_classname}", exc_info=True)
                    key_data = load_binary_file(keyfile_path)
                    if key_data and key_data.find(b"BEGIN OPENSSH PRIVATE KEY") >= 0:
                        log(" (OpenSSH private key file)")
            if key:
                log(f"auth_publickey using {keyfile_path!r} as {pkey_classname}: {keymd5(key)}")
                try:
                    transport.auth_publickey(username, key)
                except SSHException as e:
                    auth_errors.setdefault("key", []).append(str(e))
                    log(f"key {keyfile_path!r} rejected", exc_info=True)
                    log.info(f"SSH authentication using key {keyfile_path!r} failed:")
                    log.info(f" {e}")
                    if str(e) == PARAMIKO_SESSION_LOST:
                        # no point in trying more keys
                        break
                else:
                    if transport.is_authenticated():
                        break
            else:
                log.error(f"Error: cannot load private key {keyfile_path!r}")

    def auth_none() -> None:
        log("trying none authentication")
        try:
            transport.auth_none(username)
        except SSHException as e:
            auth_errors.setdefault("none", []).append(str(e))
            log("auth_none()", exc_info=True)

    def auth_password() -> None:
        log("trying password authentication")
        try:
            transport.auth_password(username, password)
        except SSHException as authe:
            estr = getattr(authe, "message", str(authe))
            auth_errors.setdefault("password", []).append(estr)
            log("auth_password(..)", exc_info=True)
            emsgs = estr.split(";")
        else:
            emsgs = []
        if not transport.is_authenticated():
            log.info("SSH password authentication failed:")
            for emsg in emsgs:
                log.info(f" {emsg}")
            if log.is_debug_enabled() and LOG_FAILED_CREDENTIALS:
                log.info(f" invalid username {username!r} or password {password!r}")

    def auth_interactive() -> None:
        log("trying interactive authentication")
        try:
            iauthhandler = IAuthHandler(password)
            transport.auth_interactive(username, iauthhandler.handle_request, "")
        except SSHException as authe:
            estr = getattr(authe, "message", str(authe))
            auth_errors.setdefault("interactive", []).append(estr)
            log("auth_interactive(..)", exc_info=True)
            log.info("SSH password authentication failed:")
            for emsg in getattr(authe, "message", estr).split(";"):
                log.info(f" {emsg}")

    banner = transport.get_banner()
    if banner:
        log.info("SSH server banner:")
        for x in banner.splitlines():
            log.info(f" {x}")

    log(f"starting authentication, authentication methods: {auth_modes}")
    auth = list(auth_modes)
    # per the RFC we probably should do none first always and read off the supported
    # methods, however, the current code seems to work fine with OpenSSH
    while not transport.is_authenticated() and auth:
        a = auth.pop(0)
        log("auth=%s", a)
        if a == "none":
            auth_none()
        elif a == "agent":
            auth_agent()
        elif a == "key":
            auth_publickey()
        elif a == "password":
            auth_interactive()
            if not transport.is_authenticated():
                if password:
                    auth_password()
                else:
                    tries = configint("numberofpasswordprompts", PASSWORD_RETRY)
                    for _ in range(tries):
                        password = input_pass(f"please enter the SSH password for {username}@{host}")
                        if not password:
                            break
                        auth_password()
                        if transport.is_authenticated():
                            break
        else:
            log.warn(f"Warning: invalid authentication mechanism {a}")
        # detect session-lost problems:
        # (no point in continuing without a session)
        if auth_errors and not transport.is_authenticated():
            for err_strs in auth_errors.values():
                if PARAMIKO_SESSION_LOST in err_strs:
                    raise SSHAuthenticationError(host, auth_errors)
    if not transport.is_authenticated():
        transport.close()
        log(f"authentication errors: {auth_errors}")
        raise SSHAuthenticationError(host, auth_errors)
    return transport


class SSHAuthenticationError(InitExit):
    def __init__(self, host, errors):
        super().__init__(ExitCode.CONNECTION_FAILED, f"SSH Authentication failed for {host!r}")
        self.errors = errors


def run_test_command(transport, cmd: str) -> tuple[list[str], list[str], int]:
    log(f"run_test_command(transport, {cmd})")
    try:
        chan = transport.open_session(window_size=None, max_packet_size=0, timeout=60)
        chan.set_name(f"run-test:{cmd}")
    except SSHException as e:
        log("open_session", exc_info=True)
        raise InitExit(ExitCode.SSH_FAILURE, f"failed to open SSH session: {e}") from None
    chan.exec_command(cmd)
    log(f"exec_command({cmd!r}) returned")
    start = monotonic()
    while not chan.exit_status_ready():
        if monotonic() - start > TEST_COMMAND_TIMEOUT:
            chan.close()
            raise InitException(f"SSH test command {cmd!r} timed out")
        log("exit status is not ready yet, sleeping")
        sleep(0.01)
    code = chan.recv_exit_status()
    log(f"exec_command({cmd!r})={code}")

    def chan_read(fileobj) -> list[str]:
        try:
            return fileobj.readlines()
        except OSError:
            log(f"chan_read({fileobj})", exc_info=True)
            return []
        finally:
            noerr(fileobj.close)

    # don't wait too long for the data:
    chan.settimeout(EXEC_STDOUT_TIMEOUT)
    out = chan_read(chan.makefile())
    log(f"exec_command out={out!r}")
    chan.settimeout(EXEC_STDERR_TIMEOUT)
    err = chan_read(chan.makefile_stderr())
    log(f"exec_command err={err!r}")
    chan.close()
    return out, err, code


def run_remote_xpra(transport, xpra_proxy_command=None, remote_xpra=None,
                    socket_dir=None, display_as_args=None, paramiko_config=None):
    log("run_remote_xpra%s", (transport, xpra_proxy_command, remote_xpra,
                              socket_dir, display_as_args, paramiko_config))
    assert remote_xpra
    log(f"will try to run xpra from: {remote_xpra}")

    def rtc(cmd: str) -> tuple[list[str], list[str], int]:
        return run_test_command(transport, cmd)

    def detectosname() -> str:
        # first, try a syntax that should work with any ssh server:
        r = rtc("echo %OS%")
        if r[2] == 0 and r[0]:
            name = r[0][-1].rstrip("\n\r")
            log(f"echo %OS%={name!r}")
            if name != "%OS%":
                # MS Windows OS will return "Windows_NT" here
                log.info(f"ssh server OS is {name!r}")
                return name
        # this should work on all other OSes:
        r = rtc("echo $OSTYPE")
        if r[2] == 0 and r[0]:
            name = r[0][-1].rstrip("\n\r")
            log(f"OSTYPE={name!r}")
            log.info(f"ssh server OS is {name!r}")
            return name
        return "unknown"

    def getexeinstallpath():
        cmd = WIN32_REGISTRY_QUERY
        if osname == "msys":
            # escape for msys shell:
            cmd = cmd.replace("/", "//")
        r = rtc(cmd)
        if r[2] != 0:
            return None
        out = r[0]
        for line in out:
            qmatch = re.search(r"InstallPath\s*\w*\s*(.*)", line)
            if qmatch:
                return qmatch.group(1).rstrip("\n\r")
        return None

    osname = detectosname()
    tried = set()
    find_command = None
    for xpra_cmd in remote_xpra:
        found = False
        if osname.startswith("Windows") or osname in ("msys", "cygwin"):
            # on MS Windows,
            # always prefer the application path found in the registry:
            def winpath(p) -> str:
                if osname == "msys":
                    return p.replace("\\", "\\\\")
                if osname == "cygwin":
                    return "/cygdrive/" + p.replace(":\\", "/").replace("\\", "/")
                return p

            installpath = getexeinstallpath()
            if installpath:
                xpra_cmd = winpath(f"{installpath}\\Xpra_cmd.exe")
                found = True
            elif xpra_cmd.find("/") < 0 and xpra_cmd.find("\\") < 0:
                test_path = winpath(f"{DEFAULT_WIN32_INSTALL_PATH}\\{xpra_cmd}")
                cmd = f'dir "{test_path}"'
                r = rtc(cmd)
                if r[2] == 0:
                    xpra_cmd = test_path
                    found = True
        if not found and not find_command and not osname.startswith("Windows"):
            if rtc("command")[2] == 0:
                find_command = "command -v"
            else:
                find_command = "which"
        if not found and find_command:
            r = rtc(f"{find_command} {xpra_cmd}")
            out = r[0]
            if r[2] == 0 and out:
                # use the actual path returned by 'command -v' or 'which':
                try:
                    xpra_cmd = out[-1].rstrip("\n\r ").lstrip("\t ")
                except Exception as e:
                    log(f"cannot get command from {xpra_cmd}: {e}")
                else:
                    if xpra_cmd.startswith(f"alias {xpra_cmd}="):
                        # ie: "alias xpra='xpra -d proxy'" -> "xpra -d proxy"
                        xpra_cmd = xpra_cmd.split("=", 1)[1].strip("'")
                    found = bool(xpra_cmd)
            elif xpra_cmd == "xpra" and osname in ("msys", "cygwin"):
                default_path = CYGWIN_DEFAULT_PATH if osname == "cygwin" else MSYS_DEFAULT_PATH
                if default_path:
                    # try the default system installation path
                    r = rtc(f"command -v '{default_path}'")
                    if r[2] == 0:
                        xpra_cmd = default_path  # ie: "/mingw64/bin/xpra"
                        found = True
        if not found or xpra_cmd in tried:
            continue
        log(f"adding {xpra_cmd=!r}")
        tried.add(xpra_cmd)
        cmd = '"' + xpra_cmd + '" ' + ' '.join(shellquote(x) for x in xpra_proxy_command)
        if socket_dir:
            cmd += f" \"--socket-dir={socket_dir}\""
        if display_as_args:
            cmd += " "
            cmd += " ".join(shellquote(x) for x in display_as_args)
        log(f"cmd({xpra_proxy_command}, {display_as_args})={cmd}")

        # see https://github.com/paramiko/paramiko/issues/175
        # WINDOW_SIZE = 2097152
        log(f"trying to open SSH session, window-size={WINDOW_SIZE}, timeout={TIMEOUT}")
        try:
            chan = transport.open_session(window_size=WINDOW_SIZE, max_packet_size=0, timeout=TIMEOUT)
            chan.set_name("run-xpra")
        except SSHException as e:
            log("open_session", exc_info=True)
            raise InitExit(ExitCode.SSH_FAILURE, f"failed to open SSH session: {e}") from None
        agent_option = str((paramiko_config or {}).get("agent", SSH_AGENT)) or "no"
        log(f"paramiko {agent_option=}")
        if agent_option.lower() in TRUE_OPTIONS:
            if not WIN_OPENSSH_AGENT and WIN_OPENSSH_AGENT_MODULE not in sys.modules:
                log(f"preventing {WIN_OPENSSH_AGENT_MODULE!r} from loading")
                sys.modules[WIN_OPENSSH_AGENT_MODULE] = None
            log.info("paramiko SSH agent forwarding enabled")
            from paramiko.agent import AgentRequestHandler
            AgentRequestHandler(chan)
        log(f"channel exec_command({cmd!r})")
        chan.exec_command(cmd)
        log("exec_command sent, returning channel for service")
        return chan
    raise RuntimeError("all SSH remote proxy commands have failed - is xpra installed on the remote host?")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# This file is part of Xpra.
# Copyright (C) 2020 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import unittest

from xpra.os_util import WIN32
from xpra.util.env import nomodule_context
from xpra.scripts.parsing import (
    parse_ssh_option, get_ssh_args, get_ssh_proxy_args, parse_remote_display,
)


class TestParsing(unittest.TestCase):

    def test_ssh_parsing(self):
        assert parse_ssh_option("auto")[0] in ("paramiko", "ssh")
        assert parse_ssh_option("ssh") == ["ssh"]
        assert parse_ssh_option("ssh -v") == ["ssh", "-v"]
        with nomodule_context("paramiko"):
            def pssh(s, e):
                r = parse_ssh_option(s)[0]
                assert r == e, f"expected {e} got {r}"

            if WIN32:
                pssh("auto", "plink.exe")
            else:
                pssh("auto", "ssh")

        # args:
        def targs(e, *args):
            r = get_ssh_args(*args)
            assert r == e, f"expected {e} but got {r}"

        targs([], {"host": "host"})
        targs(["-pw", "password1", "-l", "username1", "-P", "2222", "-T", "host1"], {
            "username": "username1",
            "password": "password1",
            "host": "host1",
            "port": 2222,
        }, ["putty"])
        if not WIN32:
            keyfile = os.path.expanduser("~/key")
            targs(["-l", "username1", "-p", "2222", "-T", "host1", "-i", keyfile], {
                "username": "username1",
                "password": "password1",
                "host": "host1",
                "port": 2222,
                "key": keyfile,
            }, ["ssh"])

        # ssh proxy:
        def pargs(e, n, *args):
            r = get_ssh_proxy_args(*args)[:n]
            assert r == e, f"expected {e} but got {r}"

        pargs(["-o"], 1, {
            "proxy_username": "username1",
            "proxy_password": "password1",
            "proxy_host": "host1",
            "proxy_port": 2222,
        }, ["ssh"])
        pargs(["-proxycmd"], 1, {
            "proxy_username": "username1",
            "proxy_password": "password1",
            "proxy_host": "host1",
            "proxy_port": 2222,
        }, ["plink"])
        # remote display attributes:
        assert parse_remote_display("somedisplay").get("display") == "somedisplay"
        assert parse_remote_display("10?proxy=username:password@host:222").get("proxy") == "username:password@host:222"

        def t(s):
            v = parse_remote_display(s)
            assert v.get("display") == "somedisplay"

        t("somedisplay?proxy=")
        t("somedisplay?proxy=:22")
        t("somedisplay?proxy=:@host:22")
        t("somedisplay?proxy=:password@host:22")


def main():
    unittest.main()


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import sys
from typing import Any
import notify2

from xpra.notifications.notifier_base import NotifierBase, NID


class PyNotify_Notifier(NotifierBase):
    CACHE: dict[int, Any] = {}

    def show_notify(self, dbus_id, tray, nid: NID,
                    app_name: str, replaces_nid: NID, app_icon,
                    summary: str, body: str, actions, hints, timeout: int, icon) -> None:
        if not self.dbus_check(dbus_id):
            return
        icon_string = self.get_icon_string(nid, app_icon, icon)
        if not notify2.is_initted():
            notify2.init(app_name or "Xpra", "glib")
        n = notify2.Notification(summary, body, icon_string)
        PyNotify_Notifier.CACHE[int(nid)] = n
        n.set_urgency(notify2.URGENCY_LOW)
        n.set_timeout(timeout)
        n.show()
        if icon_string:
            def notification_closed(*_args):
                self.clean_notification(int(nid))

            n.connect("closed", notification_closed)

    def clean_notification(self, nid: int) -> None:
        PyNotify_Notifier.CACHE.pop(nid, None)
        super().clean_notification(nid)

    def close_notify(self, nid: NID) -> None:
        n = PyNotify_Notifier.CACHE.pop(int(nid), None)
        if n:
            n.close()


def main(args):
    from xpra.os_util import gi_import
    Gtk = gi_import("Gtk")
    GLib = gi_import("GLib")
    summary = "Summary"
    body = "Body..."
    if len(args) > 1:
        summary = args[1]
    if len(args) > 2:
        body = args[2]

    def show():
        nid = 1
        n = PyNotify_Notifier()
        n.show_notify("", None, nid, "Test", 0, "", summary, body, ["0", "Hello", "1", "Bye"], {}, 0, "")
        GLib.timeout_add(5000, n.close_notify, nid)
        return False

    GLib.idle_add(show)
    GLib.timeout_add(20000, Gtk.main_quit)
    Gtk.main()


if __name__ == "__main__":
    main(sys.argv)



# This file is part of Xpra.
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
from typing import Any
from collections.abc import Callable

from xpra.os_util import gi_import
from xpra.log import Logger

log = Logger("x11", "window")
metalog = Logger("x11", "window", "metadata")

GObject = gi_import("GObject")

PROPERTIES_DEBUG = [x.strip() for x in os.environ.get("XPRA_WINDOW_PROPERTIES_DEBUG", "").split(",")]


class AutoPropGObjectMixin:
    """Mixin for automagic property support in GObjects.

    Make sure this is the first entry on your parent list, so super().__init__
    will work right."""

    def __init__(self):
        self._gproperties: dict[str, Any] = {}

    def do_get_property(self, pspec):
        return self._gproperties.get(pspec.name)

    def do_set_property(self, pspec, value) -> None:
        self._internal_set_property(pspec.name, value)

    def _internal_set_property(self, name: str, value) -> None:
        if name in PROPERTIES_DEBUG:
            metalog.info("_internal_set_property(%r, %r)", name, value, backtrace=True)
        self._gproperties[name] = value
        self.notify(name)


class WindowModelStub(AutoPropGObjectMixin, GObject.GObject):
    """
        Stub for all window models
    """

    # things that we expose:
    _property_names: list[str] = []
    # exposed and changing (should be watched for notify signals):
    _dynamic_property_names: list[str] = []
    _internal_property_names: list[str] = []
    _MODELTYPE: str = "Stub"

    def __init__(self):
        AutoPropGObjectMixin.__init__(self)
        GObject.GObject.__init__(self)
        self._setup_done: bool = False  # so we can ignore notify() events during setup
        self._managed: bool = False
        self._managed_handlers: list[int] = []

    #########################################
    # Setup and teardown
    #########################################

    def is_managed(self) -> bool:
        return self._managed

    def unmanage(self, _exiting=False) -> None:
        self.managed_disconnect()

    #########################################
    # Connect to signals in a "managed" way
    #########################################

    def managed_connect(self, detailed_signal, handler, *args) -> int:
        """ connects a signal handler and makes sure we will clean it up on unmanage() """
        handler_id: int = self.connect(detailed_signal, handler, *args)
        self._managed_handlers.append(handler_id)
        return handler_id

    def managed_disconnect(self) -> None:
        for handler_id in self._managed_handlers:
            self.disconnect(handler_id)
        self._managed_handlers = []

    ################################
    # Property reading
    ################################

    def get_dimensions(self):
        return NotImplementedError()

    #########################################
    # Properties we choose to expose
    #########################################

    def get_property_names(self) -> list[str]:
        """ The properties that should be exposed to clients """
        return self._property_names

    def get_dynamic_property_names(self) -> list[str]:
        """ The properties that may change over time """
        return self._dynamic_property_names

    def get_internal_property_names(self) -> list[str]:
        """ The properties that should not be exposed to the client """
        return self._internal_property_names

    @staticmethod
    def get_logger(property_name) -> Callable:
        if property_name in PROPERTIES_DEBUG:
            return metalog.info
        return metalog.debug

    def _updateprop(self, name: str, value) -> bool:
        """ Updates the property and fires notify(),
            but only if the value has changed
            and if the window has finished setting up, and it is still managed.
            Can only be used for AutoPropGObjectMixin properties.
        """
        logger = self.get_logger(name)
        cur = self._gproperties.get(name, None)
        if name not in self._gproperties or cur != value:
            logger("updateprop(%s, %s) previous value=%s", name, value, cur)
            self._gproperties[name] = value
            if self._setup_done and self._managed:
                self.notify(name)
            else:
                logger("not sending notify(%s) (setup done=%s, managed=%s)", name, self._setup_done, self._managed)
            return True
        logger("updateprop(%s, %s) unchanged", name, value)
        return False

    def get(self, name: str, default_value=None) -> object:
        """ Allows us to avoid defining all the attributes we may ever query,
            returns the default value if the property does not exist.
        """
        logger = self.get_logger(name)
        if name in set(self._property_names + self._dynamic_property_names + self._internal_property_names):
            v = self.get_property(name)
            logger("get(%s, %s) using get_property=%s", name, default_value, v)
        else:
            v = default_value
            if name not in ("override-redirect", "tray"):
                logger("get(%s, %s) not a property of %s, returning default value=%s", name, default_value, type(self), v)
        return v

    def show(self) -> None:
        """
        implemented for real windows only,
        trays and OR windows are always "shown" somewhere
        """

    # temporary? / convenience access methods:
    def is_OR(self) -> bool:
        """ Is this an override-redirect window? """
        return bool(self.get("override-redirect", False))

    def is_tray(self) -> bool:
        """ Is this a tray window? """
        return bool(self.get("tray", False))

    def is_shadow(self) -> bool:
        """ Is this a shadow instead of a real window? """
        return False

    def has_alpha(self) -> bool:
        """ Does the pixel data have an alpha channel? """
        return bool(self.get("has-alpha", False))



# This file is part of Xpra.
# Copyright (C) 2022 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from xpra.os_util import gi_import
from xpra.scripts.config import InitExit
from xpra.exit_codes import ExitCode
from xpra.x11.server.shadow import ShadowX11Server
from xpra.server.shadow.root_window_model import RootWindowModel
from xpra.gtk.util import get_default_root_window
from xpra.codecs.image import ImageWrapper
from xpra.codecs.evdi.capture import EvdiDevice, find_evdi_devices, add_evdi_device  # pylint: disable=no-name-in-module
from xpra.log import Logger

GObject = gi_import("GObject")
GLib = gi_import("GLib")

log = Logger("server")
log.enable_debug()


class EVDIModel(RootWindowModel):
    def __repr__(self):
        return f"EVDIModel({self.capture} : {self.geometry})"

    def get_image(self, x, y, width, height) -> ImageWrapper | None:
        dev = self.capture.evdi_device
        log.warn(f"get_image({x}, {y}, {width}, {height}) using {self.capture}, device={dev}")
        if not dev:
            return None
        # r = dev.refresh()
        # log.warn(f"refresh()={r}")
        ld = self.capture.last_damage
        if not ld:
            return None
        bw, bh, buf = ld[:3]
        if bw != width or bh != height:
            return None
        return ImageWrapper(0, 0, width, height, buf, "BGRX", 24, width * 4, )


class ExpandServer(GObject.GObject, ShadowX11Server):

    def __init__(self, attrs: dict[str, str]):
        GObject.GObject.__init__(self)
        ShadowX11Server.__init__(self, attrs)
        self.session_type = "expand"
        self.evdi_device = None
        self.evdi_channel = None
        self.fd_source = None
        self.fd_watch = None
        self.last_damage: tuple = ()

    def init(self, opts) -> None:
        ShadowX11Server.init(self, opts)
        # pylint: disable=import-outside-toplevel
        from xpra.codecs.evdi.load import load_evdi_module
        if not load_evdi_module():
            log.warn("Warning: ensure that the 'evdi' kernel module is loaded")
        devices = find_evdi_devices()
        if not devices:
            raise InitExit(ExitCode.DEVICE_NOT_FOUND, "no evdi devices found")

    def no_windows(self) -> None:
        pass
        # self.cancel_refresh_timer()
        # self.cancel_poll_pointer()

    def refresh(self) -> bool:
        # we have to continue to call the device refresh,
        # otherwise the UI hangs completely!
        dev = self.evdi_device
        log(f"refresh() using device {dev}")
        if dev:
            dev.refresh()
        # return dev is None
        return True

    def evdi_io_event(self, channel, condition) -> bool:
        log.warn(f"io_event({channel}, {condition})")
        self.evdi_device.handle_events()
        # self.evdi_device.refresh()
        return True

    def evdi_setup(self) -> None:
        # import time
        # time.sleep(2)
        log("evdi_setup()")
        devices = find_evdi_devices()
        if not devices:
            # try to add one:
            add_evdi_device()
            devices = find_evdi_devices()
        device = -1 if not devices else devices[0]
        log("devices=%s, device=%i", devices, device)
        dev = EvdiDevice(device, self.evdi_damage)
        self.evdi_device = dev
        log(f"evdi_setup() {dev=}")
        dev.open()
        dev.connect()
        dev.enable_cursor_events()
        log("evdi_setup() done")

    def start_evdi_watch(self) -> bool:
        self.evdi_setup()
        # self.evdi_device.handle_all_events()
        self.fd_source = self.evdi_device.get_event_fd()
        self.evdi_channel = GLib.IOChannel.unix_new(fd=self.fd_source)
        self.evdi_channel.set_encoding(None)
        self.evdi_channel.set_buffered(False)
        self.evdi_channel.set_close_on_unref(True)
        self.fd_watch = GLib.io_add_watch(self.evdi_channel, GLib.PRIORITY_LOW, GLib.IO_IN, self.evdi_io_event)
        return False

    def do_run(self) -> None:
        self.start_refresh_timer()
        # GLib.timeout_add(1*1000, self.start_evdi_watch)
        self.start_evdi_watch()
        super().do_run()

    def evdi_damage(self, width, height, buf, rects) -> None:
        log("evdi_damage(%s)", (width, height, buf, rects))
        self.last_damage = width, height, buf, rects
        self.refresh_windows()

    def cleanup(self) -> None:
        fdw = self.fd_watch
        if fdw:
            self.fd_watch = None
            GLib.source_remove(fdw)
        c = self.evdi_channel
        if c:
            self.evdi_channel = None
            c.shutdown(False)
        ed = self.evdi_device
        if ed:
            self.evdi_device = None
            ed.cleanup()
        super().cleanup()

    def sanity_checks(self, _proto, c) -> bool:
        return True

    def start_poll_pointer(self) -> None:
        """ not needed """

    def get_server_mode(self) -> str:
        return "X11 expand"

    def set_refresh_delay(self, v: int) -> None:
        assert 0 < v < 10000
        self.refresh_delay = v

    def setup_capture(self):
        return None

    def get_root_window_model_class(self) -> type:
        return EVDIModel

    def verify_capture(self, ss) -> None:
        pass

    def makeRootWindowModels(self) -> list[EVDIModel]:
        # TODO: remove root window
        root = get_default_root_window()
        geom = (0, 0, 800, 600)
        model = EVDIModel(root, self, "evdi", geom)
        return [model]

    def do_make_screenshot_packet(self):
        raise NotImplementedError()


GObject.type_register(ExpandServer)



#!/usr/bin/env python3

import sys


def dump_xshape(xid):
    from xpra.x11.bindings.window import X11WindowBindings, SHAPE_KIND  #@UnresolvedImport
    X11Window = X11WindowBindings()
    extents = X11Window.XShapeQueryExtents(xid)
    if not extents:
        print("read_shape for window %#x: no extents" % xid)
        return {}
    v = {}
    bextents = extents[0]
    cextents = extents[1]
    if bextents[0] == 0 and cextents[0] == 0:
        print("read_shape for window %#x: none enabled" % xid)
        return {}
    v["Bounding.extents"] = bextents
    v["Clip.extents"] = cextents
    for kind in SHAPE_KIND.keys():
        kind_name = SHAPE_KIND[kind]
        rectangles = X11Window.XShapeGetRectangles(xid, kind)
        v[kind_name + ".rectangles"] = rectangles
    print("read_shape()=%s" % v)
    return v


def main(args):
    from xpra.x11.gtk.display_source import init_gdk_display_source
    init_gdk_display_source()
    for wid in args[1:]:
        print("looking for window %s" % wid)
        if wid.startswith("0x"):
            dump_xshape(int(wid[2:], 16))
        else:
            dump_xshape(int(wid))


if __name__ == '__main__':
    sys.exit(main(sys.argv))



# This file is part of Xpra.
# Copyright (C) 2018 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import signal
import subprocess
from typing import Iterable
from collections.abc import Callable

from xpra.gtk.window import add_close_accel, add_window_accel
from xpra.gtk.widget import imagebutton, label
from xpra.gtk.pixbuf import get_icon_pixbuf
from xpra.gtk.dialogs.util import hb_button, nearest_icon_size
from xpra.os_util import gi_import, WIN32
from xpra.util.env import IgnoreWarningsContext
from xpra.exit_codes import exit_str
from xpra.common import NotificationID, noop
from xpra.platform.paths import get_xpra_command
from xpra.log import Logger

Gtk = gi_import("Gtk")
Gdk = gi_import("Gdk")
GLib = gi_import("GLib")
Gio = gi_import("Gio")

log = Logger("util")


def exec_command(cmd: list[str]) -> subprocess.Popen:
    env = os.environ.copy()
    env["XPRA_WAIT_FOR_INPUT"] = "0"
    proc = subprocess.Popen(cmd, env=env)
    log("exec_command(%s)=%s", cmd, proc)
    return proc


class BaseGUIWindow(Gtk.Window):

    def __init__(self,
                 title="Xpra",
                 icon_name="xpra.png",
                 wm_class=("xpra-gui", "Xpra-GUI"),
                 default_size=(640, 300),
                 header_bar=(True, True, False),
                 parent: Gtk.Window | None = None,
                 ):
        self.exit_code = 0
        super().__init__()
        if header_bar:
            self.add_headerbar(*header_bar)
        self.set_title(title)
        self.set_border_width(10)
        self.set_resizable(True)
        self.set_decorated(True)
        self.set_position(Gtk.WindowPosition.CENTER)
        self.icon_name = icon_name
        icon = get_icon_pixbuf(icon_name)
        if icon:
            self.set_icon(icon)
        if parent:
            if not WIN32:
                self.set_transient_for(parent)
            self.do_dismiss = self.hide
        else:
            self.do_dismiss = self.quit
        self.connect("delete_event", self.dismiss)
        add_close_accel(self, self.dismiss)
        add_window_accel(self, 'F1', self.show_about)
        with IgnoreWarningsContext():
            self.set_wmclass(*wm_class)
        self.vbox = Gtk.VBox(homogeneous=False, spacing=10)
        self.set_box_margin()
        self.vbox.set_vexpand(True)
        self.add(self.vbox)
        self.populate()
        self.vbox.show_all()
        self.set_default_size(*default_size)
        self.connect("focus-in-event", self.focus_in)
        self.connect("focus-out-event", self.focus_out)

    def set_box_margin(self, start=40, end=40, top=0, bottom=20) -> None:
        self.vbox.set_margin_start(start)
        self.vbox.set_margin_end(end)
        self.vbox.set_margin_top(top)
        self.vbox.set_margin_bottom(bottom)

    def clear_vbox(self) -> None:
        for x in self.vbox.get_children():
            self.vbox.remove(x)

    def populate_form(self, lines: Iterable[str] = (), *buttons) -> None:
        self.clear_vbox()
        self.add_widget(label(self.get_title(), font="sans 20"))
        self.add_text_lines(lines)
        self.add_buttons(*buttons)

    def add_text_lines(self, lines, font="Sans 14") -> None:
        text = "\n".join(lines)
        lbl = label(text, font=font)
        lbl.set_line_wrap(True)
        lbl.set_use_markup(True)
        self.add_widget(lbl)

    def add_buttons(self, *buttons) -> list[Gtk.Button]:
        hbox = Gtk.HBox()
        hbox.set_vexpand(False)
        self.add_widget(hbox)
        btnlist: list[Gtk.Button] = []
        for button_label, callback in buttons:
            btn = Gtk.Button.new_with_label(button_label)
            btn.connect("clicked", callback)
            btnlist.append(btn)
            hbox.pack_start(btn, True, True)
        self.show_all()
        return btnlist

    def dismiss(self, *args) -> None:
        log(f"dismiss{args} calling {self.do_dismiss}")
        self.do_dismiss()

    def add_headerbar(self, about=True, toolbox=True, configure=False) -> None:
        log("add_headerbar%s", (about, toolbox, configure))
        hb = Gtk.HeaderBar()
        hb.set_show_close_button(True)
        hb.props.title = "Xpra"

        # fixup the icon size when the window headerbar is bigger than expected:
        fixed_size = [0]
        hb_buttons = []

        def add_hb_button(*args) -> None:
            size = fixed_size[0] or 32
            btn = hb_button(*args, size=size)
            hb_buttons.append(btn)
            hb.add(btn)

        def add_gui(text: str, icon_name: str, gui_class) -> None:
            def show_gui(*_args) -> None:
                w = None

                def hide(*_args) -> None:
                    w.hide()

                gui_class.quit = hide
                w = gui_class()
                w.show()
            add_hb_button(text, icon_name, show_gui)

        def add_buttons() -> None:
            if about:
                add_hb_button("About", "help-about", self.show_about)
            if toolbox:
                try:
                    from xpra.gtk.dialogs.toolbox import ToolboxGUI
                except ImportError:
                    pass
                else:
                    add_gui("Toolbox", "toolbox", ToolboxGUI)
            if configure:
                try:
                    from xpra.gtk.configure.main import ConfigureGUI
                except ImportError:
                    pass
                else:
                    add_gui("Configure", "features", ConfigureGUI)
            hb.show_all()

        def rs(rect) -> tuple[int, int]:
            return rect.width, rect.height

        def fix_default_icon_sizes(*args) -> None:
            log("fix_default_icon_sizes%s", args)
            changed = []

            def fix_button_size(btn: Gtk.Button) -> None:
                btn_size = rs(btn.get_allocated_size()[0])
                for child in btn.get_children():
                    if not isinstance(child, Gtk.Image):
                        continue
                    size = nearest_icon_size(max(24, btn_size[1] - 4))
                    child.set_size_request(size, size)
                    btn.set_size_request(size, size)
                    if abs(fixed_size[0] - size) >= 8:
                        changed.append((fixed_size[0], size, btn, child))
                    fixed_size[0] = size

            def with_child(child) -> None:
                if not isinstance(child, Gtk.Box):
                    return
                for g_child in child.get_children():
                    if isinstance(g_child, Gtk.Button):
                        fix_button_size(g_child)
            hb.forall(with_child)

        def map_event(*_args) -> None:
            GLib.timeout_add(100, fix_default_icon_sizes)
        self.connect("map-event", map_event)

        add_buttons()
        self.set_titlebar(hb)

    def ib(self, title="", icon_name="browse.png", tooltip="", callback: Callable = noop, sensitive=True) -> Gtk.Button:
        label_font = "sans 16"
        icon = get_icon_pixbuf(icon_name)
        btn = imagebutton(
            title=title, icon=icon,
            tooltip=tooltip, clicked_callback=callback,
            icon_size=48, label_font=label_font,
        )
        btn.set_sensitive(sensitive)
        self.add_widget(btn)
        return btn

    def add_widget(self, widget) -> None:
        self.vbox.add(widget)

    # noinspection PyMethodMayBeStatic
    def focus_in(self, window, event) -> None:
        log("focus_in(%s, %s)", window, event)

    def focus_out(self, window, event) -> None:
        log("focus_out(%s, %s)", window, event)
        self.reset_cursors()

    def app_signal(self, signum: int | signal.Signals) -> None:
        if self.exit_code is None:
            self.exit_code = 128 + int(signum)
        log("app_signal(%s) exit_code=%i", signum, self.exit_code)
        GLib.idle_add(self.quit)

    def hide(self, *args) -> None:
        log("hide%s", args)
        super().hide()

    def quit(self, *args) -> None:
        log("quit%s", args)
        self.do_quit()

    # noinspection PyMethodMayBeStatic
    def do_quit(self) -> None:
        log("do_quit()")
        Gtk.main_quit()

    def show_about(self, *_args) -> None:
        from xpra.gtk.dialogs.about import about
        about(parent=self)

    def get_xpra_command(self, *args) -> list[str]:
        return get_xpra_command() + list(args)

    def button_command(self, btn, *args) -> None:
        cmd = self.get_xpra_command(*args)
        proc = exec_command(cmd)
        if proc.poll() is None:
            self.busy_cursor(btn)
            from xpra.util.child_reaper import getChildReaper
            getChildReaper().add_process(proc, "subcommand", cmd, ignore=True, forget=True,
                                         callback=self.command_ended)

    def command_ended(self, proc) -> None:
        self.reset_cursors()
        log(f"command_ended({proc})")
        if proc.returncode:
            self.may_notify(NotificationID.FAILURE,
                            "Subcommand Failed",
                            "The subprocess terminated abnormally\n\rand returned %s" % exit_str(proc.returncode)
                            )

    def busy_cursor(self, widget) -> None:
        from xpra.gtk.cursors import cursor_types
        watch = cursor_types.get("WATCH")
        if watch:
            display = Gdk.Display.get_default()
            cursor = Gdk.Cursor.new_for_display(display, watch)
            widget.get_window().set_cursor(cursor)
            GLib.timeout_add(5 * 1000, self.reset_cursors)

    def reset_cursors(self, *_args) -> None:
        for widget in self.vbox.get_children():
            widget.get_window().set_cursor(None)

    def exec_subcommand(self, subcommand, *args) -> None:
        log("exec_subcommand(%s, %s)", subcommand, args)
        cmd = get_xpra_command()
        cmd.append(subcommand)
        cmd += list(args)
        proc = exec_command(cmd)
        if proc.poll() is None:
            self.hide()

            def may_exit() -> None:
                if proc.poll() is None:
                    self.quit()
                else:
                    self.show()

            # don't ask me why,
            # but on macos we can get file descriptor errors
            # if we exit immediately after we spawn the `attach` command
            GLib.timeout_add(2000, may_exit)

    def may_notify(self, nid: NotificationID, summary: str, body: str) -> None:
        log.info(summary)
        log.info(body)
        from xpra.platform.gui import get_native_notifier_classes
        nc = get_native_notifier_classes()
        if not nc:
            return
        from xpra.util.objects import make_instance
        notifier = make_instance(nc)
        if not notifier:
            return
        from xpra.platform.paths import get_icon_filename
        from xpra.notifications.common import parse_image_path
        icon_filename = get_icon_filename(self.icon_name)
        icon = parse_image_path(icon_filename)
        notifier.show_notify(0, None, nid,
                             "xpra GUI Window", 0, self.icon_name,
                             summary, body, {}, {}, 10, icon)



# This file is part of Xpra.
# Copyright (C) 2012 Antoine Martin <antoine@xpra.org>
# Copyright (C) 2010 Nathaniel Smith <njs@pobox.com>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import signal
import threading
from collections.abc import Callable

from xpra.common import noop
from xpra.exit_codes import ExitValue
from xpra.net.bytestreams import untilConcludes
from xpra.util.str_fn import repr_ellipsized, hexstr
from xpra.util.env import envint, envbool
from xpra.os_util import force_quit, POSIX
from xpra.log import Logger

log = Logger("proxy")

SHOW_DATA = envbool("XPRA_PROXY_SHOW_DATA")
PROXY_BUFFER_SIZE = envint("XPRA_PROXY_BUFFER_SIZE", 65536)


def noretry(_e) -> bool:
    return False


class XpraProxy:
    """
        This is the proxy command that runs
        when one uses the hidden subcommand
        "xpra _proxy".
        It simply forwards stdin/stdout to
        the server socket.
    """

    def __repr__(self):
        return f"XpraProxy({self._name}: {self._client_conn} - {self._server_conn})"

    def __init__(self, name, client_conn, server_conn, quit_cb: Callable | None = None):
        self._name = name
        self._client_conn = client_conn
        self._server_conn = server_conn
        self._quit_cb: Callable = quit_cb or self.do_quit
        self._closed = False
        self._to_client = threading.Thread(target=self._to_client_loop, daemon=True)
        self._to_server = threading.Thread(target=self._to_server_loop, daemon=True)
        self.exit_code = 0
        signal.signal(signal.SIGINT, self.signal_quit)
        signal.signal(signal.SIGTERM, self.signal_quit)
        if POSIX:
            signal.signal(signal.SIGHUP, self.signal_quit)
            signal.signal(signal.SIGPIPE, self.signal_quit)

    def start_threads(self) -> None:
        self._to_client.start()
        self._to_server.start()

    def run(self) -> ExitValue:
        log("XpraProxy.run() %s", self._name)
        self.start_threads()
        self._to_client.join()
        self._to_server.join()
        log("XpraProxy.run() %s: all the threads have ended, calling quit() to close the connections", self._name)
        self.quit()
        return self.exit_code

    def _to_client_loop(self) -> None:
        self._copy_loop(f"<-server {self._name}", self._server_conn, self._client_conn)
        self._closed = True

    def _to_server_loop(self) -> None:
        self._copy_loop(f"->server {self._name}", self._client_conn, self._server_conn)
        self._closed = True

    def _copy_loop(self, log_name: str, from_conn, to_conn) -> None:
        # log("XpraProxy._copy_loop(%s, %s, %s)", log_name, from_conn, to_conn)
        try:
            while not self._closed:
                log("%s: waiting for data", log_name)
                buf = untilConcludes(self.is_active, noretry, from_conn.read, PROXY_BUFFER_SIZE)
                if not buf:
                    log("%s: connection lost", log_name)
                    return
                if SHOW_DATA:
                    log("%s: %s bytes: %s", log_name, len(buf), repr_ellipsized(buf))
                    log("%s:           %s", log_name, repr_ellipsized(hexstr(buf)))
                while buf and not self._closed:
                    log("%s: writing %s bytes", log_name, len(buf))
                    written = untilConcludes(self.is_active, noretry, to_conn.write, buf)
                    buf = buf[written:]
                    log("%s: written %s bytes", log_name, written)
            log("%s copy loop ended", log_name)
        except OSError:
            log("%s", log_name, exc_info=True)
        finally:
            self.quit()

    def is_active(self) -> bool:
        return not self._closed

    def signal_quit(self, signum, _frame=None) -> None:
        self.exit_code = 128 + signum
        self.quit()

    def quit(self, *args) -> None:
        log("XpraProxy.quit(%s) %s: closing connections", args, self._name)
        self._closed = True
        try:
            self._client_conn.close()
        except OSError:
            pass
        try:
            self._server_conn.close()
        except OSError:
            pass
        quit_cb: Callable = self._quit_cb
        self._quit_cb = noop
        quit_cb(self)

    def do_quit(self, _proxy) -> None:
        force_quit(self.exit_code)



# This file is part of Xpra.
# Copyright (C) 2020 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from xpra.os_util import gi_import
from xpra.client.gtk3.menu_helper import MenuHelper
from xpra.platform.gui import get_icon_size
from xpra.log import Logger

log = Logger("gtk", "window")

Gtk = gi_import("Gtk")
Gdk = gi_import("Gdk")


def ticked_icon(state: bool) -> str:
    return "ticked.png" if state else "unticked.png"


class WindowMenuHelper(MenuHelper):

    def __init__(self, client, window):
        super().__init__(client)
        self.window = window

    def get_menu_items(self) -> list[Gtk.ImageMenuItem | Gtk.MenuItem]:
        return [
            # menu.append(self.make_aboutmenuitem())
            self.make_infomenuitem(),
            # if self.client.client_supports_opengl:
            #    menu.append(self.make_openglmenuitem())
            self.make_minimizemenuitem(),
            self.make_maximizemenuitem(),
            self.make_fullscreenmenuitem(),
            self.make_abovenmenuitem(),
            self.make_grabmenuitem(),
            self.make_bordermenuitem(),
            self.make_refreshmenuitem(),
            self.make_reinitmenuitem(),
            self.make_closemenuitem(),
        ]

    def make_infomenuitem(self) -> Gtk.ImageMenuItem:
        def show_info(*_args) -> None:
            from xpra.client.gtk3.window_info import WindowInfo
            wi = WindowInfo(self.client, self.window)
            wi.show()

        gl = self.menuitem("Window Information", "information.png", "Window state and details", show_info)
        gl.set_tooltip_text()
        return gl

    def make_openglmenuitem(self) -> Gtk.ImageMenuItem:
        gl = self.checkitem("OpenGL")
        gl.set_tooltip_text("hardware accelerated rendering using OpenGL")
        return gl

    def make_minimizemenuitem(self) -> Gtk.ImageMenuItem:
        def minimize(*args) -> None:
            log("minimize%s", args)
            self.window.iconify()

        return self.menuitem("Minimize", "minimize.png", cb=minimize)

    def make_maximizemenuitem(self) -> Gtk.ImageMenuItem:
        def maximize(*args) -> None:
            log("maximize%s", args)
            if self.window.is_maximized():
                self.window.unmaximize()
            else:
                self.window.maximize()

        def get_label(maximized) -> str:
            return "Unmaximize" if maximized else "Maximize"

        label = get_label(self.window.is_maximized())
        self.maximize_menuitem = self.menuitem(label, "maximize.png", cb=maximize)

        def window_state_updated(widget, event) -> None:
            maximized_changed = event.changed_mask & Gdk.WindowState.MAXIMIZED
            log("state_changed%s maximized_changed=%s", (widget, event), maximized_changed)
            if maximized_changed:
                lbl = get_label(event.new_window_state & Gdk.WindowState.MAXIMIZED)
                self.maximize_menuitem.set_label(lbl)
            self.maximize_menuitem.set_sensitive(self.window.can_maximize())

        self.window.connect("window-state-event", window_state_updated)
        return self.maximize_menuitem

    def make_fullscreenmenuitem(self) -> Gtk.ImageMenuItem:
        def fullscreen(*args) -> None:
            log("fullscreen%s", args)
            self.window.fullscreen()

        return self.menuitem("Fullscreen", "scaling.png", cb=fullscreen)

    def _set_ticked(self, item, state: bool) -> None:
        icon_name = ticked_icon(state)
        icon_size = self.menu_icon_size or get_icon_size()
        image = self.get_image(icon_name, icon_size)
        item.set_image(image)

    def make_abovenmenuitem(self) -> Gtk.ImageMenuItem:
        def toggle_above(*args) -> None:
            above = not self.window._above
            log("toggle_above%s above=%s", args, above)
            self.window._above = above
            self.window.set_keep_above(above)
            self._set_ticked(self.above_menuitem, above)

        self.above_menuitem = self.menuitem("Always on top", ticked_icon(self.window._above), cb=toggle_above)
        return self.above_menuitem

    def make_grabmenuitem(self) -> Gtk.ImageMenuItem:
        def is_grabbed() -> bool:
            return self.client._window_with_grab == self.window.wid

        def toggle_grab(*args) -> None:
            log("toggle_grab%s", args)
            if is_grabbed():
                self.client.window_ungrab()
            else:
                self.client.window_grab(self.window.wid, self.window)
            self._set_ticked(self.grab_menuitem, is_grabbed())

        self.grab_menuitem = self.menuitem("Grabbed", ticked_icon(is_grabbed()), cb=toggle_grab)
        return self.grab_menuitem

    def make_bordermenuitem(self) -> Gtk.ImageMenuItem:
        def toggle_border(*args) -> None:
            log("toggle_border%s", args)
            self.window.border.shown = not self.window.border.shown
            self._set_ticked(self.grab_menuitem, self.window.border.shown)
            self.client.send_refresh(self.window.wid)

        self.border_menuitem = self.menuitem("Show Border", ticked_icon(self.window.border.shown), cb=toggle_border)
        return self.border_menuitem

    def make_refreshmenuitem(self) -> Gtk.ImageMenuItem:
        def force_refresh(*args) -> None:
            log("force refresh%s", args)
            self.client.send_refresh(self.window.wid)
            reset_icon = getattr(self.window, "reset_icon", None)
            if reset_icon:
                reset_icon()

        return self.menuitem("Refresh", "retry.png", cb=force_refresh)

    def make_reinitmenuitem(self) -> Gtk.ImageMenuItem:
        def force_reinit(*args) -> None:
            log("force reinit%s", args)
            self.client.reinit_window(self.window.wid, self.window)
            reset_icon = getattr(self.window, "reset_icon", None)
            if reset_icon:
                reset_icon()

        return self.menuitem("Re-initialize", "reinitialize.png", cb=force_reinit)

    def make_closemenuitem(self) -> Gtk.ImageMenuItem:
        def close(*args) -> None:
            log("close(%s)", args)
            self.window.close()

        return self.menuitem("Close", "close.png", cb=close)



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2018 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import sys
from collections.abc import Sequence

from xpra.util.io import stderr_print
from xpra.util.objects import typedict
from xpra.auth.sys_auth_base import SysAuthenticatorBase, log
from xpra.auth.common import parse_uid, parse_gid
from xpra.net.digest import get_salt, get_digests, gendigest


class Authenticator(SysAuthenticatorBase):
    CLIENT_USERNAME = True

    def __init__(self, **kwargs):
        self.service = kwargs.pop("service", "")
        self.uid = parse_uid(kwargs.pop("uid", None))
        self.gid = parse_gid(kwargs.pop("gid", None))
        kwargs["prompt"] = kwargs.pop("prompt", "GSS token")
        super().__init__(**kwargs)
        log("gss auth: service=%r, username=%r", self.service, kwargs.get("username"))

    def get_uid(self) -> int:
        return self.uid

    def get_gid(self) -> int:
        return self.gid

    def __repr__(self):
        return "gss"

    def get_challenge(self, digests: Sequence[str]) -> tuple[bytes, str]:
        assert not self.challenge_sent
        if "gss" not in digests:
            log.error("Error: client does not support gss authentication")
            return b"", ""
        self.salt = get_salt()
        self.challenge_sent = True
        return self.salt, f"gss:{self.service}"

    def check(self, token: bytes) -> bool:
        log("check(%s)", repr(token))
        assert self.challenge_sent
        try:
            # pylint: disable=import-outside-toplevel
            from gssapi import creds as gsscreds
            from gssapi import sec_contexts as gssctx
        except ImportError as e:
            log("check(..)", exc_info=True)
            log.warn("Warning: cannot use gss authentication:")
            log.warn(" %s", e)
            return False
        server_creds = gsscreds.Credentials(usage='accept')
        server_ctx = gssctx.SecurityContext(creds=server_creds)
        server_ctx.step(token)
        return server_ctx.complete


def main(argv):
    # pylint: disable=import-outside-toplevel
    from xpra.platform import program_context
    with program_context("GSS-Auth", "GSS-Authentication"):
        if len(argv) != 3:
            stderr_print("%s invalid arguments" % argv[0])
            stderr_print("usage: %s username token" % argv[0])
            return 1
        username = argv[1]
        token = argv[2]
        kwargs = {"username": username}
        a = Authenticator(**kwargs)
        server_salt, digest = a.get_challenge(("gss", ))
        salt_digest = a.choose_salt_digest(get_digests())
        assert digest.startswith("gss:"), "unexpected digest %r" % digest
        client_salt = get_salt(len(server_salt))
        combined_salt = gendigest(salt_digest, client_salt, server_salt)
        response = gendigest(digest, token, combined_salt)
        caps = typedict({
            "challenge_response": response,
            "challenge_client_salt": client_salt,
        })
        r = a.authenticate(caps)
        print("success: %s" % bool(r))
        return r


if __name__ == "__main__":
    sys.exit(main(sys.argv))



# This file is part of Xpra.
# Copyright (C) 2019 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from xpra.util.str_fn import csv
from xpra.gtk.error import xswallow

from xpra.x11.bindings.window import X11WindowBindings
from xpra.x11.bindings.res import ResBindings

X11Window = X11WindowBindings()
XRes = ResBindings()
if not XRes.check_xres():
    XRes = None


def get_wintitle(xid: int) -> str:
    with xswallow:
        data = X11Window.XGetWindowProperty(xid, "WM_NAME", "STRING")
        if data:
            return data.decode("latin1")
    with xswallow:
        data = X11Window.XGetWindowProperty(xid, "_NET_WM_NAME", "UTF8_STRING")
        if data:
            return data.decode("utf8")
    return ""


def get_wininfo(xid: int) -> str:
    wininfo = [f"xid={xid:x}"]
    if XRes:
        with xswallow:
            pid = XRes.get_pid(xid)
            if pid:
                wininfo.append(f"pid={pid}")
    title = get_wintitle(xid)
    if title:
        wininfo.insert(0, repr(title))
        return csv(wininfo)
    while xid:
        title = get_wintitle(xid)
        if title:
            wininfo.append(f"child of {title!r}")
            return csv(wininfo)
        with xswallow:
            xid = X11Window.getParent(xid)
    return csv(wininfo)



# This file is part of Xpra.
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

REINIT_WINDOWS = True

# we access the GUI when running as a server (tray, etc.)
# and so we cannot daemonize
CAN_DAEMONIZE = False

CLIPBOARDS = ("CLIPBOARD",)
CLIPBOARD_WANT_TARGETS = True
CLIPBOARD_GREEDY = True

OPEN_COMMAND = ("open",)

# DEFAULT_SSH_COMMAND = "ssh"

COMMAND_SIGNALS = ("SIGINT", "SIGTERM", "SIGHUP", "SIGKILL", "SIGUSR1", "SIGUSR2")



# This file is part of Xpra.
# Copyright (C) 2017 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import sys
import socket
import subprocess

from xpra.common import SocketState
from xpra.platform.paths import get_xpra_command, get_nodock_command
from xpra.platform.dotxpra import DotXpra
from xpra.platform.gui import force_focus
from xpra.util.child_reaper import getChildReaper
from xpra.exit_codes import exit_str
from xpra.scripts.config import OPTION_TYPES
from xpra.scripts.main import get_command_args
from xpra.gtk.window import add_close_accel
from xpra.gtk.widget import scaled_image, imagebutton, label, modify_fg, color_parse
from xpra.gtk.pixbuf import get_icon_pixbuf
from xpra.gtk.signals import register_os_signals
from xpra.gtk.dialogs.util import hb_button
from xpra.net.common import DEFAULT_PORTS
from xpra.exit_codes import ExitCode, ExitValue
from xpra.util.objects import typedict
from xpra.os_util import gi_import, WIN32, getuid, getgid
from xpra.util.env import IgnoreWarningsContext
from xpra.log import Logger

Gtk = gi_import("Gtk")
GLib = gi_import("GLib")
Gio = gi_import("Gio")

log = Logger("client", "util")

try:
    local_host_name = socket.gethostname()
except OSError:
    local_host_name = "localhost"


def get_session_info(sockpath: str) -> dict[str, str]:
    # the lazy way using a subprocess
    if WIN32:
        socktype = "named-pipe"
    else:
        socktype = "socket"
    cmd = get_nodock_command() + ["id", f"{socktype}:{sockpath}"]
    p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
    stdout = p.communicate()[0]
    log("get_sessions_info(%s) returncode(%s)=%s", sockpath, cmd, p.returncode)
    if p.returncode != 0:
        return {}
    info: dict[str, str] = {}
    for line in stdout.splitlines():
        parts = line.split("=", 1)
        if len(parts) == 2:
            info[parts[0]] = parts[1]
    log("get_sessions_info(%s)=%s", sockpath, info)
    return info


def get_uri(password: str, interface, protocol, name: str, stype: str, domain, host: str, address, port: int, text) -> str:
    log("get_uri%s", (password, interface, protocol, name, stype, domain, host, address, port, text))
    dstr = ""
    tt = typedict(text)
    display = tt.strget("display")
    username = tt.strget("username")
    mode = tt.strget("mode")
    if not mode:
        # guess the mode from the service name,
        # ie: "localhost.localdomain :2 (wss)" -> "wss"
        # ie: "localhost.localdomain :2 (ssh-2)" -> "ssh"
        pos = name.rfind("(")
        if name.endswith(")") and pos > 0:
            mode = name[pos + 1:-1].split("-")[0]
            if mode not in ("tcp", "ws", "wss", "ssl", "ssh"):
                return ""
        else:
            mode = "tcp"
    if display and display.startswith(":"):
        dstr = display[1:]
    # append interface to IPv6 host URI for link local addresses ("fe80:"):
    if interface and address.lower().startswith("fe80:"):
        # ie: "fe80::c1:ac45:7351:ea69%eth1"
        try:
            address += "%" + socket.if_indextoname(interface)
        except OSError:
            pass
    uri = f"{mode}://"
    if username:
        if password:
            uri += f"{username}:{password}@{address}"
        else:
            uri += f"{username}@{address}"
    else:
        uri += address
    if port > 0:
        if DEFAULT_PORTS.get(mode, 0) != port:  # NOSONAR @SuppressWarnings("python:S1066")
            uri += f":{port}"
    if protocol not in ("socket", "named-pipe"):
        uri += "/"
        if dstr:
            uri += "%s" % dstr
    return uri


def get_platform_icon_name(platform: str) -> str:
    for p, i in {
        "win32": "win32",
        "darwin": "osx",
        "linux": "linux",
        "freebsd": "freebsd",
    }.items():
        if platform.startswith(p):
            return i
    return ""


class SessionsGUI(Gtk.Window):

    def __init__(self, options, title="Xpra Session Browser"):
        super().__init__()
        self.options = options
        self.exit_code = ExitCode.OK
        self.set_title(title)
        self.set_border_width(20)
        self.set_resizable(True)
        self.set_default_size(800, 220)
        self.set_decorated(True)
        self.set_size_request(800, 220)
        self.set_position(Gtk.WindowPosition.CENTER)
        with IgnoreWarningsContext():
            self.set_wmclass("xpra-sessions-gui", "Xpra-Sessions-GUI")
        add_close_accel(self, self.quit)
        self.connect("delete_event", self.quit)
        icon = get_icon_pixbuf("browse.png")
        if icon:
            self.set_icon(icon)

        hb = Gtk.HeaderBar()
        hb.set_show_close_button(True)
        hb.props.title = "Xpra"
        hb.add(hb_button("About", "help-about", self.show_about))
        hb.show_all()
        self.set_titlebar(hb)

        self.clients = {}
        self.clients_disconnecting = set()
        self.child_reaper = getChildReaper()

        self.vbox = Gtk.VBox(homogeneous=False, spacing=20)
        self.add(self.vbox)

        title_label = label(title, font="sans 14")
        title_label.show()
        self.vbox.add(title_label)

        self.warning = label(" ")
        modify_fg(self.warning, color_parse("red"))
        self.warning.show()
        self.vbox.add(self.warning)

        self.password_box = Gtk.HBox(homogeneous=False, spacing=10)
        self.password_label = label("Password:")
        al = Gtk.Alignment(xalign=1, yalign=0.5)
        al.add(self.password_label)
        al.show()
        self.password_box.add(al)
        self.password_entry = Gtk.Entry()
        self.password_entry.set_max_length(128)
        self.password_entry.set_width_chars(16)
        self.password_entry.set_visibility(False)
        al = Gtk.Alignment(xalign=0, yalign=0.5)
        al.add(self.password_entry)
        al.show()
        self.password_box.add(al)
        self.vbox.add(self.password_box)

        self.contents = None
        self.records = []
        try:
            from xpra.platform.info import get_username
            username = get_username()
        except OSError:
            username = ""
        self.local_info_cache = {}
        self.dotxpra = DotXpra(options.socket_dir, options.socket_dirs, username)
        self.poll_local_sessions()
        self.populate()
        GLib.timeout_add(5 * 1000, self.update)
        self.vbox.show()
        self.show()

    def show(self) -> None:  # pylint: disable=arguments-differ
        super().show()

        def show() -> None:
            force_focus()
            self.present()

        GLib.idle_add(show)

    def quit(self, *args) -> None:
        log("quit%s", args)
        GLib.idle_add(self.do_quit)

    def do_quit(self) -> None:
        log("do_quit()")
        self.cleanup()
        Gtk.main_quit()

    def app_signal(self, signum) -> None:
        self.exit_code = 128 + signum
        log("app_signal(%s) exit_code=%i", signum, self.exit_code)
        self.do_quit()

    def cleanup(self) -> None:
        self.close()

    def show_about(self, *_args) -> None:
        from xpra.gtk.dialogs.about import about
        about(parent=self)

    def update(self) -> bool:
        if self.poll_local_sessions():
            self.populate()
        return True

    def populate(self) -> None:
        self.populate_table()

    def poll_local_sessions(self) -> bool:
        # TODO: run in a thread so we don't block the UI thread!
        d = self.dotxpra.socket_details(matching_state=SocketState.LIVE)
        log("poll_local_sessions() socket_details=%s", d)
        info_cache = {}
        for d, details in d.items():
            log("poll_local_sessions() %s : %s", d, details)
            for state, display, sockpath in details:
                assert state == SocketState.LIVE
                key = (display, sockpath)
                info = self.local_info_cache.get(key)
                if not info:
                    # try to query it
                    try:
                        info = get_session_info(sockpath)
                        if not info:
                            log(" no data for '%s'", sockpath)
                            continue
                        if info.get("session-type") == "client":
                            log(" skipped client socket '%s': %s", sockpath, info)
                            continue
                    except Exception as e:
                        log("get_session_info(%s)", sockpath, exc_info=True)
                        log.error("Error querying session info for %s", sockpath)
                        log.estr(e)
                        del e
                    if not info:
                        continue
                info_cache[key] = info
        if WIN32:
            socktype = "named-pipe"
        else:
            socktype = "socket"

        def make_text(info: dict) -> dict:
            text = {"mode": socktype}
            for k, name in {
                "platform": "platform",
                "uuid": "uuid",
                "display": "display",
                "session-type": "type",
                "session-name": "name",
            }.items():
                v = info.get(k)
                if v is not None:
                    text[name] = v
            return text

        # first remove any records that are no longer found:
        for key in self.local_info_cache:
            if key not in info_cache:
                display, sockpath = key
                self.records = [(interface, protocol, name, stype, domain, host, address, port, text) for
                                (interface, protocol, name, stype, domain, host, address, port, text) in self.records
                                if (protocol != "socket" or domain != "local" or address != sockpath)]
        # add the new ones:
        for key, info in info_cache.items():
            if key not in self.local_info_cache:
                display, sockpath = key
                self.records.append(("", "socket", "", "", "local", socket.gethostname(), sockpath, 0, make_text(info)))
        log("poll_local_sessions() info_cache=%s", info_cache)
        changed = self.local_info_cache != info_cache
        self.local_info_cache = info_cache
        return changed

    def populate_table(self) -> None:
        log("populate_table: %i records", len(self.records))
        if self.contents:
            self.vbox.remove(self.contents)
            self.contents = None
        if not self.records:
            self.contents = label("No sessions found")
            self.vbox.add(self.contents)
            self.contents.show()
            self.set_size_request(440, 200)
            self.password_box.hide()
            return
        self.password_box.show()
        self.set_size_request(-1, -1)
        grid = Gtk.Grid()

        def l(s="") -> Gtk.Label:  # noqa: E743
            widget = label(s)
            widget.set_margin_start(5)
            widget.set_margin_end(5)
            return widget

        for i, text in enumerate(("Host", "Display", "Name", "Platform", "Type", "URI", "Connect", "Open in Browser")):
            grid.attach(l(text), i, 1, 1, 1)
        # group them by uuid
        d = {}
        session_names = {}
        address = ""
        port = 0
        row = 2
        for i, record in enumerate(self.records):
            interface, protocol, name, stype, domain, host, address, port, text = record
            td = typedict(text)
            log("populate_table: record[%i]=%s", i, record)
            uuid = td.strget("uuid")
            display = td.strget("display")
            if domain == "local" and host.endswith(".local"):
                host = host[:-len(".local")]
            if uuid:
                key = uuid
            else:
                key = (host.rstrip("."), display)
            log("populate_table: key[%i]=%s", i, key)
            d.setdefault(key, []).append((interface, protocol, name, stype, domain, host, address, port, text))
            td = typedict(text)
            session_name = td.strget("name")
            if session_name:
                session_names[key] = session_name
        for key, recs in d.items():
            if isinstance(key, tuple):
                host, display = key
                uuid = str(display)
                title = f"{host} : {display}"
            else:
                display = ""
                uuid = str(key)
                # try to find a valid host name:
                hosts = [rec[5] for rec in recs if not rec[5].startswith("local")]
                if not hosts:
                    hosts = [rec[5] for rec in recs]
                host = hosts[0]
                title = str(host)
            platform, dtype = None, None
            for rec in recs:
                td = typedict(rec[-1])
                if not platform:
                    platform = td.strget("platform")
                if not dtype:
                    dtype = td.strget("type")
                if not display:
                    display = td.strget("display")
            if title in ("localhost", "localhost.localdomain", "127.0.0.1", "::1", local_host_name):
                title = "local"
            host_label = l(title)
            if uuid != title:
                host_label.set_tooltip_text(uuid)
            # try to use an icon for the platform:
            platform_icon_name = get_platform_icon_name(platform)
            pwidget = None
            if platform_icon_name:
                pwidget = scaled_image(get_icon_pixbuf("%s.png" % platform_icon_name), 28)
                if pwidget:
                    pwidget.set_tooltip_text(platform_icon_name)
            if not pwidget:
                pwidget = l(platform)
            w, c, b = self.make_connect_widgets(key, recs, address, port, display)
            session_name = session_names.get(key, "")
            widgets = host_label, l(display), l(session_name), pwidget, l(dtype), w, c, b
            for x, widget in enumerate(widgets):
                grid.attach(widget, x, row, 1, 1)
            row += 1
        grid.show_all()
        self.contents = grid
        self.vbox.add(grid)

    def attach(self, key, uri: str) -> None:
        self.warning.set_text("")
        # preserve ssl command line arguments
        option_types = {k: v for k, v in OPTION_TYPES.items() if k.startswith("ssl")}
        cmd = get_xpra_command() + ["attach", uri] + get_command_args(
            self.options,
            uid=getuid(), gid=getgid(),
            option_types=option_types,
            cmdline=sys.argv,
        )
        env = os.environ.copy()
        env["XPRA_NOTTY"] = "1"
        proc = subprocess.Popen(cmd, env=env)
        log("attach() Popen(%s)=%s", cmd, proc)

        def proc_exit(*args) -> None:
            log("proc_exit%s", args)
            c = proc.poll()
            if key in self.clients_disconnecting:
                self.clients_disconnecting.remove(key)
            elif c not in (0, None):
                self.warning.set_text(exit_str(c).replace("_", " "))
            client_proc = self.clients.pop(key, None)
            if client_proc:
                def update() -> None:
                    self.update()
                    self.populate()

                GLib.idle_add(update)

        self.child_reaper.add_process(proc, "client-%s" % uri, cmd, True, True, proc_exit)
        self.clients[key] = proc
        self.populate()

    def browser_open(self, rec) -> None:
        import webbrowser
        password = self.password_entry.get_text()
        url = get_uri(password, *rec)
        if url.startswith("wss"):
            url = "https" + url[3:]
        else:
            assert url.startswith("ws")
            url = "http" + url[2:]
        # trim end of URL:
        #  http://192.168.1.7:10000/10 -> http://192.168.1.7:10000/
        url = url[:url.rfind("/")]
        webbrowser.open_new_tab(url)

    def make_connect_widgets(self, key, recs, address, port: int, display) -> tuple:
        d = {}
        proc = self.clients.get(key)
        if proc and proc.poll() is None:
            icon = get_icon_pixbuf("disconnected.png")

            def disconnect_client(btn) -> None:
                log("disconnect_client(%s) proc=%s", btn, proc)
                self.clients_disconnecting.add(key)
                proc.terminate()
                self.populate()

            btn = imagebutton("Disconnect", icon, clicked_callback=disconnect_client)
            return label("Already connected with pid=%i" % proc.pid), btn, label("")

        icon = get_icon_pixbuf("browser.png")
        bopen = imagebutton("Open", icon)

        icon = get_icon_pixbuf("connect.png")
        if len(recs) == 1:
            # single record, single uri:
            rec = recs[0]
            uri = get_uri("", *rec)
            bopen.set_sensitive(uri.startswith("ws"))

            def browser_open(*_args) -> None:
                self.browser_open(rec)

            bopen.connect("clicked", browser_open)
            d[uri] = rec

            def clicked(*_args) -> None:
                password = self.password_entry.get_text()
                uri = get_uri(password, *rec)
                self.attach(key, uri)

            btn = imagebutton("Connect", icon, clicked_callback=clicked)
            return label(uri), btn, bopen

        # multiple modes / uris
        uri_menu = Gtk.ComboBoxText()
        uri_menu.set_size_request(340, 48)
        # sort by protocol so TCP comes first
        order = {"socket": 0, "ssh": 1, "tcp": 2, "ssl": 3, "ws": 4, "wss": 8}
        if WIN32:
            # on MS Windows, prefer ssh which has a GUI for accepting keys
            # and entering the password:
            order["ssh"] = 0

        def cmp_key(v) -> str:
            text = v[-1]  # the text record
            mode = (text or {}).get("mode", "")
            host = v[6]
            host_len = len(host)
            # log("cmp_key(%s) text=%s, mode=%s, host=%s, host_len=%s", v, text, mode, host, host_len)
            # prefer order (from mode), then shorter host string:
            return "%s-%s" % (order.get(mode, mode), host_len)

        srecs = sorted(recs, key=cmp_key)
        has_ws = False
        for rec in srecs:
            uri = get_uri("", *rec)
            uri_menu.append_text(uri)
            d[uri] = rec
            if uri.startswith("ws"):
                has_ws = True

        def connect(*_args) -> None:
            uri = uri_menu.get_active_text()
            rec = d[uri]
            password = self.password_entry.get_text()
            uri = get_uri(password, *rec)
            self.attach(key, uri)

        uri_menu.set_active(0)
        btn = imagebutton("Connect", icon, clicked_callback=connect)

        def uri_changed(*_args) -> None:
            uri = uri_menu.get_active_text()
            ws = uri.startswith("ws")
            bopen.set_sensitive(ws)
            if ws:
                bopen.set_tooltip_text("")
            elif not has_ws:
                bopen.set_tooltip_text("no 'ws' or 'wss' URIs found")
            else:
                bopen.set_tooltip_text("select a 'ws' or 'wss' URI")

        uri_menu.connect("changed", uri_changed)
        uri_changed()

        def browser_open_option(*_args) -> None:
            uri = uri_menu.get_active_text()
            rec = d[uri]
            self.browser_open(rec)

        bopen.connect("clicked", browser_open_option)
        return uri_menu, btn, bopen


def do_main(opts) -> ExitValue:
    from xpra.platform import program_context
    from xpra.log import enable_color
    with program_context("Xpra-Session-Browser", "Xpra Session Browser"):
        enable_color()
        gui = SessionsGUI(opts)
        register_os_signals(gui.app_signal)
        Gtk.main()
        log("do_main() gui.exit_code=%i", gui.exit_code)
        return gui.exit_code


def main() -> ExitValue:  # pragma: no cover
    from xpra.scripts.config import make_defaults_struct
    opts = make_defaults_struct()
    return do_main(opts)


if __name__ == "__main__":  # pragma: no cover
    r = main()
    sys.exit(r)



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2016 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

# @PydevCodeAnalysisIgnore

import time
import errno
from ctypes import addressof, byref, c_ulong, c_char_p, c_char, c_void_p, cast, string_at
from ctypes.wintypes import DWORD, HANDLE
from typing import Any

from xpra.util.str_fn import strtobytes, memoryview_to_bytes
from xpra.net.bytestreams import Connection
from xpra.net.common import ConnectionClosedException
from xpra.platform.win32.common import (
    CloseHandle, FormatMessageSystem,
    ERROR_PIPE_BUSY, ERROR_PIPE_NOT_CONNECTED,
    IO_ERROR_STR, ERROR_BROKEN_PIPE, ERROR_IO_PENDING,
)
from xpra.platform.win32.namedpipes.common import (
    OVERLAPPED, WAIT_STR, INVALID_HANDLE,
    INFINITE,
    CreateEventA, CreateFileA,
    ReadFile, WriteFile, SetEvent,
    DisconnectNamedPipe, FlushFileBuffers, WaitNamedPipeA,
    GetLastError, SetNamedPipeHandleState, WaitForSingleObject, GetOverlappedResult,
)
from xpra.platform.win32.constants import (
    FILE_FLAG_OVERLAPPED,
    GENERIC_READ, GENERIC_WRITE,
    OPEN_EXISTING, PIPE_READMODE_BYTE,
)

from xpra.log import Logger

log = Logger("network", "named-pipe", "win32")

BUFSIZE = 65536

CONNECTION_CLOSED_ERRORS = {
    ERROR_BROKEN_PIPE: "BROKENPIPE",
    ERROR_PIPE_NOT_CONNECTED: "PIPE_NOT_CONNECTED",
}
# some of these may be redundant or impossible to hit? (does not hurt I think)
for x in ("WSAENETDOWN", "WSAENETUNREACH", "WSAECONNABORTED", "WSAECONNRESET",
          "WSAENOTCONN", "WSAESHUTDOWN", "WSAETIMEDOUT", "WSAETIMEDOUT",
          "WSAEHOSTUNREACH", "WSAEDISCON"):
    CONNECTION_CLOSED_ERRORS[getattr(errno, x)] = x


class NamedPipeConnection(Connection):
    def __init__(self, name, pipe_handle, options):
        log("NamedPipeConnection(%r, %s, %s)", name, pipe_handle, options)
        super().__init__(name, "named-pipe", options=options)
        self.pipe_handle = pipe_handle
        # noinspection PyTypeChecker,PyCallingNonCallable
        self.read_buffer = (c_char * BUFSIZE)()
        self.read_buffer_ptr = cast(addressof(self.read_buffer), c_void_p)
        self.read_event = CreateEventA(None, True, False, None)
        self.read_overlapped = OVERLAPPED()
        self.read_overlapped.hEvent = self.read_event
        self.read_overlapped.Internal = None
        self.read_overlapped.InternalHigh = None
        self.read_overlapped.union.Pointer = None
        self.write_event = CreateEventA(None, True, False, None)
        self.write_overlapped = OVERLAPPED()
        self.write_overlapped.hEvent = self.write_event
        self.write_overlapped.Internal = None
        self.write_overlapped.InternalHigh = None
        self.write_overlapped.union.Pointer = None

    def can_retry(self, e) -> bool | str:
        code = e.args[0]
        if code == errno.WSAEWOULDBLOCK:  # @UndefinedVariable pylint: disable=no-member
            return "WSAEWOULDBLOCK"
        # convert those to a connection closed:
        closed = CONNECTION_CLOSED_ERRORS.get(code)
        if closed:
            raise ConnectionClosedException(e) from None
        return False

    def untilConcludes(self, fn, *args):
        try:
            return super().untilConcludes(fn, *args)
        except Exception as e:
            code = GetLastError()
            log("untilConcludes(%s, ) exception: %s, error code=%s", fn, e, code, exc_info=True)
            closed = CONNECTION_CLOSED_ERRORS.get(code)
            if closed:
                return None
            raise OSError("%s: %s" % (e, code)) from None

    def read(self, n):
        return self._read(self._pipe_read, n)

    def _pipe_read(self, n):
        read = c_ulong(0)
        n_bytes = DWORD(min(n, BUFSIZE))
        r = ReadFile(self.pipe_handle, self.read_buffer_ptr, n_bytes, byref(read), byref(self.read_overlapped))
        log("ReadFile(%i)=%i, len=%s", n, r, read.value)
        if not r and self.pipe_handle:
            e = GetLastError()
            if e != ERROR_IO_PENDING:
                log("ReadFile: %s", IO_ERROR_STR.get(e, e))
                if e in CONNECTION_CLOSED_ERRORS:
                    raise ConnectionClosedException(CONNECTION_CLOSED_ERRORS[e])
            r = WaitForSingleObject(self.read_event, INFINITE)
            log("WaitForSingleObject(..)=%s, len=%s", WAIT_STR.get(r, r), read.value)
            if r and self.pipe_handle:
                raise RuntimeError("failed to read from named pipe handle %s" % self.pipe_handle)
        if self.pipe_handle:
            if not GetOverlappedResult(self.pipe_handle, byref(self.read_overlapped), byref(read), False):
                e = GetLastError()
                if e in CONNECTION_CLOSED_ERRORS:
                    raise ConnectionClosedException(CONNECTION_CLOSED_ERRORS[e])
                raise RuntimeError("overlapped read failed: %s" % IO_ERROR_STR.get(e, e))
        if read.value == 0:
            data = None
        else:
            data = string_at(self.read_buffer_ptr, read.value)
        log("pipe_read: %i bytes", len(data or ""))  # , binascii.hexlify(s))
        return data

    def write(self, buf, _packet_type) -> int:
        return self._write(self._pipe_write, buf)

    def _pipe_write(self, buf):
        bbuf = memoryview_to_bytes(buf)
        size = len(bbuf)
        log("pipe_write: %i bytes", size)  # binascii.hexlify(buf))
        written = c_ulong(0)
        r = WriteFile(self.pipe_handle, c_char_p(bbuf), size, byref(written), byref(self.write_overlapped))
        log("WriteFile(..)=%s, len=%i", r, written.value)
        if not r and self.pipe_handle:
            e = GetLastError()
            if e != ERROR_IO_PENDING:
                log("WriteFile: %s", IO_ERROR_STR.get(e, e))
                if e in CONNECTION_CLOSED_ERRORS:
                    raise ConnectionClosedException(CONNECTION_CLOSED_ERRORS[e])
            r = WaitForSingleObject(self.write_event, INFINITE)
            log("WaitForSingleObject(..)=%s, len=%i", WAIT_STR.get(r, r), written.value)
            if not self.pipe_handle:
                # closed already!
                return written.value
            if r:
                raise RuntimeError("failed to write buffer to named pipe handle %s" % self.pipe_handle)
        if self.pipe_handle:
            if not GetOverlappedResult(self.pipe_handle, byref(self.write_overlapped), byref(written), False):
                e = GetLastError()
                raise RuntimeError("overlapped write failed: %s" % IO_ERROR_STR.get(e, e))
            log("pipe_write: %i bytes written", written.value)
            if self.pipe_handle:
                FlushFileBuffers(self.pipe_handle)
        # SetFilePointer(self.pipe_handle, 0, FILE_BEGIN)
        return written.value

    def close(self):
        log("%s.close()", self)
        ph = self.pipe_handle
        if not ph:
            return
        self.pipe_handle = None

        def _close_err(fn, e):
            log.error("Error: %s(%s) %s", fn, ph, e)

        def logerr(fn, *args):
            try:
                fn(*args)
            except Exception as e:
                _close_err(fn, e)

        logerr(SetEvent, self.read_event)
        logerr(SetEvent, self.write_event)
        logerr(FlushFileBuffers, ph)
        logerr(DisconnectNamedPipe, ph)
        logerr(CloseHandle, ph)

    def __repr__(self):
        return self.target

    def get_info(self) -> dict[str, Any]:
        d = super().get_info()
        d["type"] = "named-pipe"
        d["closed"] = self.pipe_handle is None
        return d


def connect_to_namedpipe(pipe_name: str, timeout=10) -> HANDLE:
    log("connect_to_namedpipe(%r, %i)", pipe_name, timeout)
    start = time.time()
    while True:
        if time.time() - start >= timeout:
            raise RuntimeError("timeout waiting for named pipe '%s'" % pipe_name)
        pipe_handle = CreateFileA(strtobytes(pipe_name), GENERIC_READ | GENERIC_WRITE,
                                  0, None, OPEN_EXISTING, FILE_FLAG_OVERLAPPED, 0)
        log("CreateFileA(%s)=%s", pipe_name, pipe_handle)
        if pipe_handle != INVALID_HANDLE:
            break
        err = GetLastError()
        log("CreateFileA(..) error=%s", err)
        if err == ERROR_PIPE_BUSY:
            if WaitNamedPipeA(pipe_name, timeout * 10000) == 0:
                raise RuntimeError("timeout waiting for named pipe '%s'" % pipe_name)
        else:
            raise RuntimeError("cannot open named pipe '%s': %s" % (pipe_name, FormatMessageSystem(err)))
    # we have a valid handle!
    dw_mode = c_ulong(PIPE_READMODE_BYTE)
    r = SetNamedPipeHandleState(pipe_handle, byref(dw_mode), None, None)
    log("SetNamedPipeHandleState(..)=%i", r)
    if not r:
        log.warn("Warning: SetNamedPipeHandleState failed")
    return pipe_handle



# This file is part of Xpra.
# Copyright (C) 2017 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from typing import Any

from xpra.os_util import gi_import
from xpra.client.gtk3.opengl.client_window import GLClientWindowBase
from xpra.platform.gl_context import GLContext

GObject = gi_import("GObject")


def check_support(force_enable=False) -> dict[str, Any]:
    import warnings  # pylint: disable=import-outside-toplevel
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message=".*g_object_get_qdata:")
        warnings.filterwarnings("ignore", message=".*g_object_set_qdata_full:")
        warnings.filterwarnings("ignore", message=".*g_object_unref:")
        return GLContext().check_support(force_enable)  # pylint: disable=not-callable


class GLClientWindow(GLClientWindowBase):

    def get_backing_class(self):
        # delay the import so we can trap warnings in gl.check
        # and use the numpy import context:
        from xpra.client.gtk3.opengl.drawing_area import GLDrawingArea
        return GLDrawingArea


GObject.type_register(GLClientWindow)
